# 
# Copyright (C) 2006 OpenWrt.org
#
# This is free software, licensed under the GNU General Public License v2.
# See /LICENSE for more information.
#

KDIR:=$(BUILD_DIR)/linux-$(KERNEL)-$(BOARD)
MODULE_NAME:=$(MODELNAME)
FW_VERSION:=V1.1.2.6
VALID_REGION:=0

ifndef FIRMWARE_REGION
  FW_REGION:=""
  VALID_REGION:=1
else
  ifeq ($(FIRMWARE_REGION),WW)
    FW_REGION:=""
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),NA)
    FW_REGION:="NA"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),GR)
    FW_REGION:="GR"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),PR)
    FW_REGION:="PR"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),RU)
    FW_REGION:="RU"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),PT)
    FW_REGION:="PT"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),KO)
    FW_REGION:="KO"
    VALID_REGION:=1
  endif
  ifeq ($(FIRMWARE_REGION),IN)
    FW_REGION:="IN"
    VALID_REGION:=1
  endif
endif

ifeq ($(VALID_REGION),0)
  FW_REGION:=""
  VALID_REGION:=1
endif

KERNEL_IMAGE_NAME=vmlinux.img
KIMAGE_PADDED=vmlinux.burn
RIMAGE_PADDED=root.bin
INFO_HEAD_LENGTH:=128
HEAD_LENGTH:=126
H_VERSION:=WNR2000
FLASH_BLOCK="\x12\x32"
KERNEL_LENGTH=786432
ROOTFS_LENGTH=2359296
#ROOTFS_LENGTH=2097152

#############################################
# AR23XX plafform temporality used
#############################################
# V0.1.0.2_1.0.3	#rlease to ST 12/08/2008 by lucas

ifeq ("$(BOARD)", "ar23xx")
  MODULE_NAME=WGR612
  FW_VERSION=V0.1.0.2_1.0.3
  H_VERSION=WGR612
  FLASH_BLOCK="\x07\x15"
  KERNEL_LENGTH=458752
  ROOTFS_LENGTH=1376256
endif
#############################################

#############################################
# RTL8196B plafform temporality used
#############################################

ifeq ("$(BOARD)", "rtl8196b")
  MODULE_NAME=JWNR2000
ifeq ($(VENDOR_MODEL_NAME),)
  FW_VERSION=V1.0.0.4_1.0.12
else
ifeq ($(VENDOR_MODEL_NAME),JWNB2100)
  FW_VERSION=V1.0.0.1_1.0.5
endif
endif
  H_VERSION=JWNR2000
  KERNEL_LENGTH=720896
  ROOTFS_LENGTH=1310720
endif
#############################################

KERNEL_ENTRY=`readelf -a $(BIN_DIR)/vmlinux.$(BOARD)|grep "Entry"|cut -d":" -f 2`

ifneq ($(CONFIG_BIG_ENDIAN),y)
JFFS2OPTS     :=  --pad --little-endian --squash
SQUASHFS_OPTS :=  -le
else
JFFS2OPTS     :=  --pad --big-endian --squash
SQUASHFS_OPTS :=  -be
endif

ifneq ($(CONFIG_TARGET_ROOTFS_INITRAMFS),y)
  ifeq ($(CONFIG_TARGET_ROOTFS_JFFS2),y)
    define Image/mkfs/jffs2
		rm -rf $(BUILD_DIR)/root/jffs	
		#$(STAGING_DIR)/bin/mkfs.jffs2 $(JFFS2OPTS) -e 0x10000 -o $(KDIR)/root.jffs2-64k -d $(BUILD_DIR)/root
		ln -sf /var/resolv.conf $(BUILD_DIR)/root/etc/resolv.conf
		$(STAGING_DIR)/bin/mkfs.jffs2 $(JFFS2OPTS) -e 0x20000 -o $(KDIR)/root.jffs2-128k -d $(BUILD_DIR)/root -D device_table.txt

		# add End-of-Filesystem markers
		#echo -ne '\xde\xad\xc0\xde' >> $(KDIR)/root.jffs2-64k
		#echo -ne '\xde\xad\xc0\xde' >> $(KDIR)/root.jffs2-128k
	
		$(call Image/Build,jffs2-64k)
		$(call Image/Build,jffs2-128k)
    endef
  endif
    
  ifeq ($(CONFIG_TARGET_ROOTFS_SQUASHFS),y)
    define Image/mkfs/squashfs
		@mkdir -p $(BUILD_DIR)/root/jffs
		rm -fr $(BUILD_DIR)/root/jffs $(BUILD_DIR)/root/rom 
		ln -sf /var/resolv.conf $(BUILD_DIR)/root/etc/resolv.conf
		cd $(BUILD_DIR)/root/etc; rm -fr config hostapd.conf host hotplug.d profile protocols sysctl.conf banner
		rm -fr $(BUILD_DIR)/root/lib/config $(BUILD_DIR)/root/lib/network $(BUILD_DIR)/root/sbin/hotplug $(BUILD_DIR)/root/sbin/ifup $(BUILD_DIR)/root/sbin/ifdown $(BUILD_DIR)/root/usr/lib/common.awk $(BUILD_DIR)/root/ipkg $(BUILD_DIR)/root/usr/lib/ipkg
		echo "$(MODULE_NAME)" > $(BUILD_DIR)/root/module_name
		echo "$(FW_VERSION)" > $(BUILD_DIR)/root/firmware_version
		echo "$(FW_REGION)" > $(BUILD_DIR)/root/firmware_region
		echo "$(H_VERSION)" > $(BUILD_DIR)/root/hardware_version
# Added by wayne for specific model		
		if [ -n $(VENDOR_MODEL_NAME) ]; then \
			if [ -z $(VENDOR_MODEL_SUB_NAME) ]; then \
				echo "$(VENDOR_MODEL_NAME)" > $(BUILD_DIR)/root/vendor_model_name; \
			else \
				echo "$(VENDOR_MODEL_NAME)-$(VENDOR_MODEL_SUB_NAME)" > $(BUILD_DIR)/root/vendor_model_name; \
			fi \
		fi
		sudo mknod $(BUILD_DIR)/root/dev/console c 5 1
		sudo mknod $(BUILD_DIR)/root/dev/null  c 1 3
		sudo mknod $(BUILD_DIR)/root/dev/zero  c 1 5
		sudo mknod $(BUILD_DIR)/root/dev/urandom  c 1 9
		sudo mknod $(BUILD_DIR)/root/dev/wd  c 10 130
		sudo mknod $(BUILD_DIR)/root/dev/ttyS0 c 4 64
#modified by dvd.chen, for DNI_MTD_PARTITION
		sudo mknod $(BUILD_DIR)/root/dev/mtdblock0 b 31 0
		sudo mknod $(BUILD_DIR)/root/dev/pot b 31 1
		sudo mknod $(BUILD_DIR)/root/dev/pppoe b 31 2
		sudo mknod $(BUILD_DIR)/root/dev/mtdblock3 b 31 3
		sudo mknod $(BUILD_DIR)/root/dev/mtdblock4 b 31 4
		sudo mknod $(BUILD_DIR)/root/dev/nvram b 31 5
		sudo mknod $(BUILD_DIR)/root/dev/mtdblock6 b 31 6
		sudo mknod $(BUILD_DIR)/root/dev/mtdblock7 b 31 7
		mkdir $(BUILD_DIR)/root/dev/mtd
		sudo mknod $(BUILD_DIR)/root/dev/mtd/0 c 90 0
		sudo mknod $(BUILD_DIR)/root/dev/mtd/0ro c 90 1
		sudo mknod $(BUILD_DIR)/root/dev/mtd/1 c 90 2
		sudo mknod $(BUILD_DIR)/root/dev/mtd/1ro c 90 3
		sudo mknod $(BUILD_DIR)/root/dev/mtd/2 c 90 4
		sudo mknod $(BUILD_DIR)/root/dev/mtd/2ro c 90 5
		sudo mknod $(BUILD_DIR)/root/dev/mtd/3 c 90 6
		sudo mknod $(BUILD_DIR)/root/dev/mtd/3ro c 90 7
		sudo mknod $(BUILD_DIR)/root/dev/mtd/4 c 90 8
		sudo mknod $(BUILD_DIR)/root/dev/mtd/4ro c 90 9
		sudo mknod $(BUILD_DIR)/root/dev/mtd/5 c 90 10
		sudo mknod $(BUILD_DIR)/root/dev/mtd/5ro c 90 11
		sudo mknod $(BUILD_DIR)/root/dev/mtd/6 c 90 12
		sudo mknod $(BUILD_DIR)/root/dev/mtd/6ro c 90 13
		sudo mknod $(BUILD_DIR)/root/dev/mtd/7 c 90 14
		sudo mknod $(BUILD_DIR)/root/dev/mtd/7ro c 90 15
		sudo mknod $(BUILD_DIR)/root/dev/ppp c 108 0
		sudo mknod $(BUILD_DIR)/root/dev/tty c 5 0
		#sudo chmod 666 $(BUILD_DIR)/root/dev/tty
		sudo mknod $(BUILD_DIR)/root/dev/ptmx c 5 2 
		#sudo chmod 666 $(BUILD_DIR)/root/dev/ptmx
		mkdir $(BUILD_DIR)/root/dev/pts
		sudo mknod $(BUILD_DIR)/root/dev/pts/0 c 136 0
		sudo mknod $(BUILD_DIR)/root/dev/pts/1 c 136 1
#added by dvd.chen for miniigd
		ln -s /var/linuxigd $(BUILD_DIR)/root/etc/
		ln -s /var/wps $(BUILD_DIR)/root/etc/simplecfg
#		rm $(BUILD_DIR)/root/usr/sbin/pppd
#		rm $(BUILD_DIR)/root/usr/sbin/ez-ipupdate
#		rm $(BUILD_DIR)/root/usr/sbin/mini_upnpd
#		rm $(BUILD_DIR)/root/usr/sbin/miniigd
#		rm $(BUILD_DIR)/root/bin/snarf
		cd $(BUILD_DIR)/root/dev; ln -sf /tmp/log/log log;
		rm -f $(KDIR)/root.squashfs
		$(STAGING_DIR)/bin/mksquashfs-lzma-rtk $(BUILD_DIR)/root $(KDIR)/root.squashfs $(SQUASHFS_OPTS)
#modified by dvd.chen, for DNI_MTD_PARTITION, our root-fs started from 768K (realtek is 896k)
#		cd $(BUILD_DIR); $(BUILD_DIR)/linux/rtkload/cvimg root linux-2.4-rtl8196b/root.squashfs ../bin/$(RIMAGE_PADDED) 120000 E0000
#		cd $(BUILD_DIR); cvimg root linux-2.4-rtl8196b/root.squashfs ../bin/$(RIMAGE_PADDED) 330000 C0000
		cd $(BUILD_DIR); cvimg root linux-2.4-rtl8196b/root.squashfs ../bin/$(RIMAGE_PADDED) 330000 D0000
		cp $(KDIR)/linux-2.4.18/rtkload/linux.bin $(BIN_DIR)
		$(call Image/Build,squashfs)
		dd bs=$(HEAD_LENGTH) if=/dev/zero count=1 of=$(BIN_DIR)/head.pad
		dd bs=512K if=/dev/zero count=1 of=$(BIN_DIR)/pad.img
		cat $(BIN_DIR)/pad.img | head -c 496 > $(BIN_DIR)/pad_t.img
		cat $(BIN_DIR)/pad_t.img $(BIN_DIR)/$(RIMAGE_PADDED) > $(BIN_DIR)/root_t.burn
		dd if=$(BIN_DIR)/root_t.burn of=$(BIN_DIR)/root.burn skip=1
		rm -rf $(BIN_DIR)/root_t.burn
# Add vendor model name "VENDOR_MODEL_NAME" if defined -- added by wayne
		if [ -z $(VENDOR_MODEL_NAME) ]; then \
			echo "device:$(MODULE_NAME)" > $(BIN_DIR)/head_info.pad; \
		else \
			if [ -z $(VENDOR_MODEL_SUB_NAME) ]; then \
				echo "device:$(VENDOR_MODEL_NAME)" > $(BIN_DIR)/head_info.pad; \
			else \
				echo "device:$(VENDOR_MODEL_NAME)-$(VENDOR_MODEL_SUB_NAME)" > $(BIN_DIR)/head_info.pad; \
			fi \
		fi
		echo "version:$(FW_VERSION)" >> $(BIN_DIR)/head_info.pad
		echo "region:$(FW_REGION)" >> $(BIN_DIR)/head_info.pad
		echo "KernelSize:$(KERNEL_LENGTH)" >> $(BIN_DIR)/head_info.pad
#		echo "RootfsSize:$(ROOTFS_LENGTH)" >> $(BIN_DIR)/head_info.pad
		ls -l $(BIN_DIR)/root.burn | awk -F ' ' '{printf "%d\n",$$$$5}' > $(BIN_DIR)/root.size
		echo -n "RootfsSize:" >> $(BIN_DIR)/head_info.pad
		cat $(BIN_DIR)/root.size >> $(BIN_DIR)/head_info.pad
		echo "InfoHeadSize:$(INFO_HEAD_LENGTH)" >> $(BIN_DIR)/head_info.pad
		cat $(BIN_DIR)/head_info.pad $(BIN_DIR)/head.pad | head -c $(HEAD_LENGTH) > $(BIN_DIR)/info.pad
		echo -n -e $(FLASH_BLOCK) >> $(BIN_DIR)/info.pad
		rm -rf $(BIN_DIR)/head_info.pad $(BIN_DIR)/head.pad $(BIN_DIR)/root.size
		cat $(BIN_DIR)/linux.bin $(BIN_DIR)/pad.img | head -c $(KERNEL_LENGTH) > $(BIN_DIR)/$(KIMAGE_PADDED)
		cat $(BIN_DIR)/info.pad $(BIN_DIR)/$(KIMAGE_PADDED)  $(BIN_DIR)/root.burn > $(BIN_DIR)/$(MODULE_NAME)-no-crc.img
#		cat $(BIN_DIR)/$(KIMAGE_PADDED)  $(BIN_DIR)/root.burn > $(BIN_DIR)/$(MODULE_NAME)-no-crc.img
# Add vendor model name "VENDOR_MODEL_NAME" if defined -- added by wayne
		if [ -z $(VENDOR_MODEL_NAME) ]; then \
			$(STAGING_DIR)/../tools/appendsum $(BIN_DIR)/$(MODULE_NAME)-no-crc.img \
					$(BIN_DIR)/$(MODULE_NAME)-$(FW_VERSION)$(FW_REGION).img; \
		else \
			$(STAGING_DIR)/../tools/appendsum $(BIN_DIR)/$(MODULE_NAME)-no-crc.img \
					$(BIN_DIR)/$(VENDOR_MODEL_NAME)-$(FW_VERSION)$(FW_REGION).img; \
		fi
		rm -rf $(BIN_DIR)/info.pad $(BIN_DIR)/pad.img $(BIN_DIR)/pad_t.img $(BIN_DIR)/root.burn $(BIN_DIR)/$(KIMAGE_PADDED) $(BIN_DIR)/$(MODULE_NAME)-no-crc.img
    endef
  endif

  ifeq ($(CONFIG_TARGET_ROOTFS_CRAMFS),y)
    define Image/mkfs/cramfs
	ln -sf /var/resolv.conf $(BUILD_DIR)/root/etc/resolv.conf
	make -C $(BUILD_DIR)/linux/scripts/cramfs mkcramfs
	$(BUILD_DIR)/linux/scripts/cramfs/mkcramfs -u 0 -g 0 $(BUILD_DIR)/root $(BUILD_DIR)/../bin/target.cramfs
	cp -f $(BUILD_DIR)/linux/arch/mips/brcm-boards/bcm947xx/compressed/vmlinuz $(BUILD_DIR)/../bin/
	#cp -f $(LINUX_SRCDIR)/pre-drivers/vmlinuz $(BUILD_DIR)/../bin/
	trx -o $(BUILD_DIR)/../bin/linux.trx $(BUILD_DIR)/../bin/vmlinuz $(BUILD_DIR)/../bin/target.cramfs
	addpattern -i $(BUILD_DIR)/../bin/linux.trx -o $(BUILD_DIR)/../bin/linux_lsys.bin
	# Pad self-booting Linux to a 64 KB boundary
	cp -f $(BUILD_DIR)/linux/arch/mips/brcm-boards/bcm947xx/compressed/zImage $(BUILD_DIR)/../bin/
    endef
  endif
    
  ifeq ($(CONFIG_TARGET_ROOTFS_TGZ),y)
    define Image/mkfs/tgz
		tar -zcf $(BIN_DIR)/openwrt-$(BOARD)-$(KERNEL)-rootfs.tgz --owner=root --group=root -C $(BUILD_DIR)/root/ .
    endef
  endif
  
  
endif


ifeq ($(CONFIG_TARGET_ROOTFS_EXT2FS),y)
  E2SIZE=$(shell echo $$(($(CONFIG_TARGET_ROOTFS_FSPART)*1024)))
  
  define Image/mkfs/ext2
		$(STAGING_DIR)/bin/genext2fs -q -b $(E2SIZE) -I 1500 -d $(BUILD_DIR)/root/ $(KDIR)/root.ext2
		$(call Image/Build,ext2)
  endef
endif


define Image/mkfs/prepare/default
	find $(BUILD_DIR)/root -type f -not -perm +0100 -not -name 'ssh_host*' | xargs chmod 0644
	find $(BUILD_DIR)/root -type f -perm +0100 | xargs chmod 0755
	find $(BUILD_DIR)/root -type d | xargs chmod 0755
	mkdir -p $(BUILD_DIR)/root/tmp
	chmod 0777 $(BUILD_DIR)/root/tmp
endef

define Image/mkfs/prepare
	$(call Image/mkfs/prepare/default)
endef

define BuildImage
compile:
	$(call Build/Compile)

install:
	$(call Image/Prepare)
	$(call Image/mkfs/prepare)
	$(call Image/BuildKernel)
#	$(call Image/mkfs/jffs2)
	$(call Image/mkfs/squashfs)
	$(call Image/mkfs/cramfs)
	$(call Image/mkfs/tgz)
	$(call Image/mkfs/ext2)
	
clean:
	$(call Build/Clean)
endef

compile-targets:
install-targets:
clean-targets:

download:
prepare:
compile: compile-targets
install: compile install-targets
clean: clean-targets
