/***********************************************************************
*
* The Rome was not built in a day, keep moving 2007! :-)
*
* PPPD plugin for kernel-mode PPTP on Linux
*
* Copyright (C) 2007 by Delta Network Inc., haiyue wang
*
* Borrows heavily from the PPPoE plugin by Michal Ostrowski and Jamal Hadi Salim.
* and PPTP-Linux-1.1.0 by C. Scott Ananian.
*
***********************************************************************/
#include "pptp.h"

#include "pppd/pppd.h"
#include "pppd/fsm.h"
#include "pppd/lcp.h"
#include "pppd/ipcp.h"
#include "pppd/pathnames.h"

#include <linux/types.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netdb.h>

extern int new_style_driver;
extern int session_ctl_fd;
extern int unsuccess;
extern int retry_time;
	
char pppd_version[] = VERSION;

char pptp_hostname[64];
char pptp_connection_id[64]; 

static int pptp_plugin_hook(char *cmd, char **argv, int doit);

static option_t pptp_options[] = {
	 /* we use 'devnam' to save our PPTP server address. */
	{ "Server Address", o_wild, (void *) &pptp_plugin_hook, "PPTP Server Address",
	   OPT_INITONLY | OPT_PRIVFIX | OPT_NOARG  | OPT_A2STRVAL | OPT_STATIC, devnam},

	{ "pptp_hostname", o_string, pptp_hostname,
      "hostname for pptp", OPT_PRIO | OPT_STATIC, NULL, 64 },

        { "pptp_connection_id", o_string, &pptp_connection_id,
      "connection id for pptp", OPT_PRIO | OPT_STATIC, NULL, 64 },
	{ NULL }
};

static struct pptp_conn_mngr *pptp_conn;

struct in_addr get_ip_address(char *name) 
{
	struct in_addr ret;
	struct hostent *host = gethostbyname(name);

	if (host == NULL || host->h_addrtype != AF_INET)
		pptp_fatal("Can't get the IP Address");
   
	memcpy(&ret.s_addr, host->h_addr, sizeof(ret.s_addr));

	return ret;
}

static void init_pptp(char * hostname)
{
	struct pptp_conn_mngr * conn;
	
	conn = malloc(sizeof(struct pptp_conn_mngr));

	if (conn == NULL)
		pptp_fatal("Can't allocate memory for PPTP session");

	memset(conn, 0, sizeof(struct pptp_conn_mngr));

	conn->svr_addr = get_ip_address(hostname);

	conn->call_socket = -1;
	conn->gre_socket = -1;

	conn->conn_state = CONN_IDLE;
	conn->call_state = CALL_IDLE;

	//conn->call_id = (u16)time(NULL);
	conn->call_sernum = 0x2312;

	pptp_conn = conn;
}

static int pptp_open_session(void)
{
	session_ctl_fd = -1;
	struct pptp_conn_mngr *conn = pptp_conn;
	struct sockaddr_pptp sp;

	//conn->call_id++;
	conn->call_sernum++;

	strcpy(ppp_devnam, "gre0");
	open_callmngr(conn);
	
	if (conn->call_state != CALL_ESTABLISHED) {
		fprintf(stderr, "Unable to complete PPTP Control-Call Connection!\n");
		return -1;
	}

	pptp_info("PPTP Call-ID: %u, Peer-Call-ID: %u", conn->call_id, 
			conn->peer_call_id);
	
	conn->gre_socket = socket(AF_INET, SOCK_GRE_PPP, IPPROTO_GRE);
	if (conn->gre_socket < 0)
		pptp_fatal("Failed to create PPTP socket: %m");
		
	sp.sa_family = AF_INET;
	sp.pa.call_id = htons(conn->call_id);
	sp.pa.peer_callid = htons(conn->peer_call_id);
	sp.pa.peer_addr = conn->svr_addr.s_addr;
	
	if (connect(conn->gre_socket, (struct sockaddr *) &sp, sizeof(struct sockaddr_pptp)) < 0)
		pptp_fatal("Failed to connect PPTP socket: %d %m", errno);

	session_ctl_fd = conn->call_socket;

	return conn->gre_socket;
}


static void pptp_send_config(int mtu, u32 asyncmap, int pcomp, int accomp)
{
	if (mtu > MAX_PPTP_MTU)
		pptp_info("Couldn't increase MTU to %d", mtu);
}


static void pptp_recv_config(int mru, u32 asyncmap, int pcomp, int accomp)
{
	if (mru > MAX_PPTP_MTU)
		pptp_info("Couldn't increase MRU to %d", mru);
}

static void pptp_close_session(void)
{
	struct pptp_conn_mngr * conn = pptp_conn;

	/* Maybe it has been disconnected by server. */
	if (conn->call_state == CALL_ESTABLISHED)
		close_callmngr(conn);
	
	close(conn->gre_socket);
	close(conn->call_socket);

	conn->gre_socket = -1;
	conn->call_socket = -1;
	
	conn->call_state = CALL_IDLE;
	conn->conn_state = CONN_IDLE;
}

static void pptp_input_packet(void)
{
	int ret = -1;
	struct pptp_conn_mngr * conn = pptp_conn;

	
	if (conn->call_state == CALL_ESTABLISHED)
		ret = pptp_get_input(conn);

	if (ret == PPTP_RET_TERM_LINK) {
		link_terminated(0);

		/* Let's set the bellowing value to connect server soon!*/
		unsuccess = 1;
	}
}

void pptp_time_out(void * arg)
{
	struct pptp_conn_mngr * conn = (struct pptp_conn_mngr *)arg;

	if (conn->call_state == CALL_ESTABLISHED) {
		pptp_handle_timer(conn);

		/* Repeat!*/
		TIMEOUT(pptp_time_out, conn, 30);
	}
}

void pptp_timer(int on)
{
	struct pptp_conn_mngr * conn = pptp_conn;

	if (on && conn->call_state == CALL_ESTABLISHED)
		TIMEOUT(pptp_time_out, conn, 30);

	/* Just try to do it!*/
	if (on == 0)
		UNTIMEOUT(pptp_time_out, conn);
}

struct channel pptp_channel = {
	.options	= pptp_options,
	.process_extra_options	= NULL,
	.check_options			= NULL,
	.connect				= &pptp_open_session,
	.disconnect			= &pptp_close_session,
	.establish_ppp			= &generic_establish_ppp,
	.disestablish_ppp		= &generic_disestablish_ppp,
	.send_config			= &pptp_send_config,
	.recv_config			= &pptp_recv_config,
	.close				= NULL,
	.cleanup				= NULL,
	.input_packet			= &pptp_input_packet,
	.handle_timer			= &pptp_timer,
};

static int pptp_plugin_hook(char *cmd, char **argv, int doit)
{
	strncpy(devnam, cmd, sizeof(devnam));
	
	if (the_channel != &pptp_channel) {
		the_channel = &pptp_channel;

		modem = 0;

		lcp_allowoptions[0].neg_accompression = 0;
		lcp_wantoptions[0].neg_accompression = 0;

		lcp_allowoptions[0].neg_asyncmap = 0;
		lcp_wantoptions[0].neg_asyncmap = 0;

		lcp_allowoptions[0].neg_pcompression = 0;
		lcp_wantoptions[0].neg_pcompression = 0;

		ipcp_allowoptions[0].neg_vj=0;
		ipcp_wantoptions[0].neg_vj=0;

		init_pptp(devnam);
	}

	return 1;
}


void plugin_init(void)
{
	if (!ppp_available() && !new_style_driver)
		pptp_fatal("Linux kernel does not support PPTP -- are you running 2.4.x?");

	add_options(pptp_options);
}

