#include <signal.h>
#include <sys/wait.h>
#include <arpa/inet.h>

#include "soap.h"

#define OS_VERSION		"Linux/2.4"
#define SOAP_SERVER_STRING	OS_VERSION " Boa/0.94 soap/1.0"

#define HOST_NAME		"JWNR2000"
#define SMARTAGENT_VERSION	"3.0"
#define SPEC_VERSION		"1.7"

#define FIREWALL_INFO		""  //need update
#define VPN_INFo		"PPP 2.4.3"

#define FW_MODULE		"/module_name"
#define SE_NUMBER		"/tmp/Seria_Number" //need update
#define FW_REGION		"/firmware_region"
#define FW_VERSION		"/firmware_version"
#define HW_VERSION		"/hardware_version"

#define SOAP_PMLIST_FILE 		"/tmp/upnp_soap_pmlist"
#define SOAP_RESULT_FILE	"/tmp/upnp_soap_result"
#define PORT_MAP_FILE 		"/tmp/upnp_portmap"

#define UNDONE			"   <ResponseCode>501</ResponseCode>\r\n"

char config_state = CONFIG_FINISHED;
char detect_state = CONFIG_FINISHED;
char detect_conntype[16] = "Down";

struct region_info
{
	char *cont_name;
	char *full_name;
};

struct timezone_info
{	
	char *GMT_value;
	char *ntp_flag;
};

int _eval(char *const argv[], char *path, int timeout, int *ppid)
{
	pid_t pid;
	int status;
	int fd;
	int flags;
	int sig;

	switch (pid = fork()) {
	case -1:        /* error */
		perror("fork");
		return errno;
	case 0:         /* child */
		/* Reset signal handlers set for parent process */
		for (sig = 0; sig < (_NSIG-1); sig++)
			signal(sig, SIG_DFL);

		/* Clean up */
		ioctl(0, TIOCNOTTY, 0);
		close(STDIN_FILENO);
		setsid();

		/* Redirect stdout to <path> */
		if (path) {
			flags = O_WRONLY | O_CREAT;
			if (!strncmp(path, ">>", 2)) {
				/* append to <path> */
				flags |= O_APPEND;
				path += 2;
  			} else if (!strncmp(path, ">", 1)) {
				 /* overwrite <path> */
				flags |= O_TRUNC;
				path += 1;
			}
			if ((fd = open(path, flags, 0644)) < 0)
				perror(path);
			else {
				dup2(fd, STDOUT_FILENO);
				close(fd);
			}
		}

		/* execute command */
		dprintf("%s\n", argv[0]);
		setenv("PATH", "/sbin:/bin:/usr/sbin:/usr/bin", 1);
		alarm(timeout);
		execvp(argv[0], argv);
		perror(argv[0]);
		exit(errno);
	default:        /* parent */
		if (ppid) {
			*ppid = pid;
			return 0;
		} else {
			if (waitpid(pid, &status, 0) == -1) {
				perror("waitpid");
				return errno;
			}
			if (WIFEXITED(status))
				return WEXITSTATUS(status);
			else
				return status;
		}
	}
}


/* Simple version of _eval() (no timeout and wait for child termination) */
#define eval(cmd, args...) ({ \
        char *argv[] = { cmd, ## args, NULL }; \
        _eval(argv, ">/dev/console", 0, NULL); \
})

#define evald(cmd, args...) ({ \
	int pid; \
	char *argv[] = { cmd, ## args, NULL }; \
	_eval(argv, ">/dev/console", 0, &pid); \
})

void do_cmd(char *cmd)
{
	if (fork() == 0) {
		system(cmd);
		exit(0);
	}
}

void do_two_cmd(char *cmd1, char *cmd2)
{
	if (fork() == 0) {
		system(cmd1);
		if (cmd2)
			system(cmd2);
		exit(0);
	}
}

void wireless_cmd(char *cmd1, char *cmd2)
{
	if (fork() == 0) {
		eval(cmd1);
		if (cmd2)
			eval(cmd2);
		exit(0);
	}
}

char *config_get(const char *name)
{
	char *buf;

	buf = nvram_get(name);
	return buf;
}

int config_set(const char *name, const char *value)
{

	nvram_set(name,value);
	return 0;
}

char *cat_file(const char *name)
{
	int i;
	FILE *fp;
	static char buf[512];

	buf[0] = '\0';

	fp = fopen(name, "r");
	if (fp == NULL)
		return buf;
	fgets(buf, sizeof(buf), fp);
	fclose(fp);

	i = 0;
	while (buf[i] != '\0' && buf[i] != '\r' && buf[i] != '\n')
		i++;
	buf[i] = '\0';

	return buf;
}

char * get_if_addr(char *name)
{
	static char buf[64];
	int fd;
	char *ret = "";
	struct ifreq ifr;

	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name,name);

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0)
		return ret;

	if (ioctl(fd, SIOCGIFADDR,&ifr) < 0)
		goto err;

	if (ifr.ifr_addr.sa_family == AF_INET) {
		struct sockaddr_in *sin = (struct sockaddr_in *)&(ifr.ifr_addr);
		strcpy(buf, inet_ntoa(sin->sin_addr));
		ret = buf;
	}

err:
	close(fd);
	return ret;
}

char *ether_etoa(const unsigned char *e, char *a)
{
	int i;
	char *c = a;

	for (i = 0; i < 6; i++) {
		if (i)
			*c++ = ':';
		c += sprintf(c, "%02X", e[i] & 0xFF);
	}
	
	return a;
}
/*
char *get_if_mac(char *name)
{
	int s;
	struct ifreq ifr;
	char eabuf[32], mac[32];
	char *ret = "";
		
	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (s < 0)
		return ret;

	strncpy(ifr.ifr_name, name, IFNAMSIZ);
	if (ioctl(s, SIOCGIFHWADDR, &ifr) == 0) {
		ret = ether_etoa(ifr.ifr_hwaddr.sa_data, eabuf);

		int i = 0, j = 0;
		while(ret[i] != '\0' && ret[i] != '\n') {
			if (ret[i] != ':')
				mac[j++] = ret[i];
			i++;
		}
		mac[j] = '\0';

		ret = mac;
	}
	close(s);

	return ret;

}*/

static void GMT2INT(char *value , char *dst)
{
	static struct timezone_info ntp[] = {
			{ "-12", "37" },			
			{ "-11", "36" },			
			{ "-10", "34" },	
			{ "-9", "33" },	
			{ "-8", "31" },
			{ "-7", "30" },
			{ "-6", "29" },
			{ "-5", "27" },
			{ "-4", "26" },
			{ "-3", "24" },
			{ "-2", "21" },
			{ "-1", "17" },
			{ "0", "16" },
			{ "+1", "14" },
			{ "+2", "13" },
			{ "+3", "12" },
			{ "+4", "11" },
			{ "+5", "9" },
			{ "+6", "7" },
			{ "+7", "6" },
			{ "+8", "4" },
			{ "+9", "3" },
			{ "+10", "2" },
			{ "+11", "1" },
			{ "+12", "0" },
	};
	int i=0;
	for (i = 0; ntp[i].GMT_value; i++) {
		if (strcmp(ntp[i].GMT_value, value) == 0) {
				strcpy(dst, ntp[i].ntp_flag);
				break;
		}
	}
	
}

void soap_send_ok(request * req, int len)
{
	req_write(req, "HTTP/1.0 200 OK\r\nCONTENT-LENGTH:");
	req_write(req, simple_itoa(len));
	req_write(req, "\r\nCONTENT-TYPE: text/xml; charset=\"UTF-8\"\r\nSERVER: ");
	req_write(req, SOAP_SERVER_STRING);
	req_write(req, "\r\n\r\n");
}

static void BuildSendAndCloseSoapResp(request * req, char * body)
{
	static char beforebody[] =
		"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
		"<SOAP-ENV:Envelope\r\n"
		"   xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\"\r\n"
		"   SOAP-ENV:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">\r\n"
		"<SOAP-ENV:Body>\r\n";

	static char afterbody[] =
		"</SOAP-ENV:Body>\r\n"
		"</SOAP-ENV:Envelope>\r\n";

	soap_send_ok(req, strlen(beforebody)+strlen(body)+strlen(afterbody));
	req_write(req, beforebody);
	req_write(req, body);
	req_write(req, afterbody);

	req_flush(req);
}

/******************** DeviceConfig Methods ********************/
static void DeviceConfig_ConfigurationStarted(request * req)
{
	static char resp[] =
		"   <m:ConfigurationStartedResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:ConfigurationStartedResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	BuildSendAndCloseSoapResp(req, resp);
	config_state = CONFIG_STARTED;
}

static void DeviceConfig_ConfigurationFinished(request * req)
{
	static char resp[] =
		"   <m:ConfigurationFinishedResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:ConfigurationFinishedResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	if (config_state == CONFIG_STARTED) {
		BuildSendAndCloseSoapResp(req, resp);
		config_state = CONFIG_FINISHED;
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetGUILanguage(request * req)
{
	static char resp[] =
		"   <m:SetGUILanguageResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetGUILanguageResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char status[8] = "", old_region[64] = "", new_region[64] = "";
	char *tmp_region;
	FILE *fp;
	int i = 0;

	static struct region_info region[] = {
			{ "CHS", "Chinese" },			//Simplified Chinese
			{ "CHT", "Chinese" },			//Traditional Chinese
			{ "Simplified Chinese", "Chinese" },	//to be deleted
			{ "Traditional Chinese", "Chinese" },	//to be delted
			{ "DEU", "German" },
			{ "ENU", "English" },
		/*
			{ "CSY", "Czech" },
			{ "DAN", "Danish" },
			{ "ELL",  "Greek" },
			{ "ESP", "Spanish" },
			{ "FIN", "Finnish" },
			{ "FRA", "French" },
			{ "HUN", "Hungarian" },
			{ "ITA", "Italian" },
			{ "JPN", "Japanese" },
			{ "KOR", "Korean" },
			{ "NLD", "Dutch" },
			{ "NOR", "Norwegian" },
			{ "PLK", "Polish" },
			{ "PTB", "Portuguese" },
			{ "ROM", "Romanian" },
			{ "RUS", "Russian" },
			{ "SKY", "Slovak" },
			{ "SLV", "Slovenian" },
			{ "SVE", "Swedish" },
		*/
			{ 0, 0 }
	};

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		tmp_region = GetValueFromNameValueList(&data, "GUILanguage");

		strcpy(new_region, tmp_region);
		for (i = 0; region[i].cont_name; i++) {
			if (strcmp(region[i].cont_name, tmp_region) == 0) {
				strcpy(new_region, region[i].full_name);
				break;
			}
		}

		for (i = 0; region[i].full_name; i++) {
			if (strcmp(region[i].full_name, new_region) == 0)
				break;
		}
		if (!region[i].full_name)
			strcpy(new_region, "English");

		if (new_region) {
			sprintf(old_region, config_get("GUI_Region"));
			if (strcmp(old_region, new_region) == 0) {
				BuildSendAndCloseSoapResp(req, resp);
				ClearNameValueList(&data);
				return;
			} else if (strcmp(new_region, "German") == 0)
				do_cmd("/sbin/cmd_traffic_meter restart");
			else
				do_cmd("/sbin/cmd_traffic_meter stop");

			config_set("GUI_Region", new_region);
			system("/www/cgi-bin/fast_test_wan_status.sh");

			if ((fp = fopen("/tmp/language_change_status", "r"))) {
				fgets(status, sizeof(status), fp);
				fclose(fp);
			}

			if (status && *status == '0')
				BuildSendAndCloseSoapResp(req, resp);
			else {
				config_set("GUI_Region", old_region);
				SoapError(req, "501");
			}
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetEnable(request * req)
{
	static char resp[] =
		"   <m:SetEnableResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetEnableResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *blank_state;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		blank_state = GetValueFromNameValueList(&data, "NewEnable");

		if (blank_state && (*blank_state == '0' || *blank_state == '1')) {	
			config_set("dns_hijack", blank_state);
			//system("/usr/sbin/nvram commit");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetConfiguration(request * req)
{
	static char resp[] =
		"   <m:SetConfigurationResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetConfigurationResponse>\r\n"
		"   <ResponseCode>001</ResponseCode>\r\n";

	if (config_state == CONFIG_STARTED)
		BuildSendAndCloseSoapResp(req, UNDONE);
	else
		SoapError(req, "501");
}

static void DeviceConfig_Loaddefault(request * req)
{
	static char resp[] =
		"   <m:LoaddefaultResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:LoaddefaultResponse>\r\n"
		"   <ResponseCode>001</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *state;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		state = GetValueFromNameValueList(&data, "NewLoaddefault");

		if (state && *state == '1') {	
			//system("/usr/sbin/nvram restore");
			system("/usr/sbin/nvram set restore_defaults=1");
			system("/usr/sbin/nvram set action=99");
			sleep(3);
			BuildSendAndCloseSoapResp(req, resp);
			//do_cmd("reboot");
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetTimeZone(request * req)
{
	static char resp[] =
		"   <m:SetTimeZoneResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetTimeZoneResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char time_zone[16], *p, *daylight, buf[16]="0";

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		p = GetValueFromNameValueList(&data, "NewTimeZone");
		daylight = GetValueFromNameValueList(&data, "NewDaylightSaving");

		if (p && daylight && (*daylight == '0' || *daylight == '1')) {
			if (*p == '+')
				*p = '-';
			else if (*p == '-')
				*p = '+';
			
			if (*p == '0')
				sprintf(time_zone, "GMT-0");
			else
				sprintf(time_zone, "GMT%s", p);
			GMT2INT(p, buf);
			config_set("ntpserver_select", time_zone);
			config_set("ntp_hidden_select", buf);
			config_set("ntpadjust", daylight);
			//do_cmd("/usr/sbin/nvram commit");
			//sleep(1);
			do_cmd("/etc/rc.d/ntpclient.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetFirmware(request * req)
{
	static char resp[] =
		"   <m:SetFirmwareResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetFirmwareResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	if (config_state == CONFIG_STARTED)
		BuildSendAndCloseSoapResp(req, UNDONE);
	else
		SoapError(req, "501");
}

static void DeviceConfig_SetBlockSiteEnable(request * req)
{
	static char resp[] =
		"   <m:SetBlockSiteEnableResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetBlockSiteEnableResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewBlockSiteEnable");

		if (enable && (*enable == '0' || *enable == '1')) {
			if (*enable == '0'){
				config_set("block_skeyword", "0");
				do_cmd("/etc/rc.d/blk_site.sh stop");
			}else{
				config_set("block_skeyword", "2");
				do_two_cmd("/etc/rc.d/blk_site.sh start", "/etc/rc.d/cmdsched.sh start site_svc");
			}
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void DeviceConfig_SetBlockSiteName(request * req)
{
	static char resp[] =
		"   <m:SetBlockSiteNameResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"   </m:SetBlockSiteNameResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char buf[1024], *p, *q;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		p = GetValueFromNameValueList(&data, "NewBlockSiteName");

		if (p) {
			
			if (*p == '1' && *(p+1) == '0' && *(p+2) == '@')
				p = p + 3;
			else if (*p >= '0' && *p <= '9' && *(p+1) == '@')
				p = p + 2;
			else {
				SoapError(req, "402");
				return;
			}

			q = p;
			while (*q != '\0' && *q != '\n') {
				if (*q == '@')
					*q = ' ';

				q++;
			}

			config_set("block_KeyWord_DomainList", p);
			if(strcmp(config_get("block_skeyword"), "2")==0)
				do_cmd("/etc/rc.d/blk_site.sh start");
			if(strcmp(config_get("block_skeyword"), "0")!=0)
				do_cmd("/etc/rc.d/cmdsched.sh start site_svc");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

void get_timezone(char *tz)
{
	char buf[16];
	
	strcpy(buf, config_get("ntpserver_select"));
	
	/* "GMT-8" --> "-8"; "GMT+12" --> "+12"; "GMT-0" --> "0" */
	if (buf[3] == '\0' || buf[3] == 'G')
		strcpy(tz, "0");
	else {
		if (buf[3] == '-')
			tz[0] = '+';
		else if (buf[3] == '+')
			tz[0] = '-';
		
		tz[1] = buf[4];
		
		if (buf[5] == 'G' || buf[5] == '\0')
			tz[2] = '\0';
		else {
			tz[2] = buf[5];
			tz[3] = '\0';
		}

		if (tz[1] == '0' && tz[2] == '\0')
			strcpy(tz, "0");
	}
}

void get_blocksite_items(char *enable, char *list)
{
	char buf[1024], *p;
	int num = 0;

	if (enable[0] != '0') {
		strcpy(enable, "1");
		sprintf(buf, config_get("block_KeyWord_DomainList"));
		
		p = buf;
		if (*p == '\0' || *p == '\n')
			return;
		
		while (*p != '\0' && *p != '\n') {
			if (*p == ' ') {
				if (num == 9) {
					*p = '\0';
					break;
				}
				
				*p = '@'; 
				num++;
			}
			p++;
		}
		
		sprintf(list, "%d@%s", num+1, buf);	
	}

}

static void DeviceConfig_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"      <BlankState>%s</BlankState>\r\n"
		"      <NewBlockSiteEnable>%s</NewBlockSiteEnable>\r\n"
		"      <NewBlockSiteName>%s</NewBlockSiteName>\r\n"
		"      <NewTimeZone>%s</NewTimeZone>\r\n"
		"      <NewDaylightSaving>%s</NewDaylightSaving>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[2048], blankstate[8];
	char siteenable[8] = "", sitename[1024] = "", time_zone[8] = "";

	sprintf(blankstate, config_get("dns_hijack"));
	sprintf(siteenable, config_get("block_skeyword"));
	
	get_timezone(time_zone);
	get_blocksite_items(siteenable, sitename);

	sprintf(body, resp, blankstate, siteenable, sitename, time_zone, 
		config_get("ntpadjust"));
	
	BuildSendAndCloseSoapResp(req, body);
}

static void DeviceConfig_GetTimeZoneInfo(request * req)
{
	static char resp[] =
		"   <m:GetTimeZoneInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"      <NewTimeZone>%s</NewTimeZone>\r\n"
		"      <NewDaylightSaving>%s</NewDaylightSaving>\r\n"
		"   </m:GetTimeZoneInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], time_zone[8] = "";

	get_timezone(time_zone);

	sprintf(body, resp, time_zone, config_get("ntpadjust"));
	BuildSendAndCloseSoapResp(req, body);
}

static void DeviceConfig_GetConfigInfo(request * req)
{
	static char resp[] =
		"   <m:GetConfigInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"      <NewConfigFile>%s</NewConfigFile>\r\n"
		"   </m:GetConfigInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024];

	BuildSendAndCloseSoapResp(req, UNDONE);
}

static void DeviceConfig_GetBlockSiteInfo(request * req)
{
	static char resp[] =
		"   <m:GetBlockSiteInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:DeviceConfig:1\">\r\n"
		"      <NewBlockSiteEnable>%s</NewBlockSiteEnable>\r\n"
		"      <NewBlockSiteName>%s</NewBlockSiteName>\r\n"
		"   </m:GetBlockSiteInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], siteenable[8] = "", sitename[1024] = "";

	sprintf(siteenable, config_get("block_skeyword"));
	
	get_blocksite_items(siteenable, sitename);
	
	sprintf(body, resp, siteenable, sitename);
	BuildSendAndCloseSoapResp(req, body);
}

static struct method_struct soapMethod1[] =
{
	{ "ConfigurationStarted",  DeviceConfig_ConfigurationStarted },
	{ "ConfigurationFinished", DeviceConfig_ConfigurationFinished },

	{ "SetGUILanguage", DeviceConfig_SetGUILanguage },	
	{ "SetEnable", DeviceConfig_SetEnable },
	{ "SetConfiguration",  DeviceConfig_SetConfiguration },
	{ "Loaddefault", DeviceConfig_Loaddefault },
	{ "SetTimeZone", DeviceConfig_SetTimeZone },
	{ "SetFirmware",  DeviceConfig_SetFirmware },
	{ "SetBlockSiteEnable", DeviceConfig_SetBlockSiteEnable },
	{ "SetBlockSiteName", DeviceConfig_SetBlockSiteName },
	//{ "SetBlockServices",  DeviceConfig_SetBlockServices },

	{ "GetInfo", DeviceConfig_GetInfo },
	{ "GetTimeZoneInfo", DeviceConfig_GetTimeZoneInfo },
	{ "GetConfigInfo", DeviceConfig_GetConfigInfo },
	{ "GetBlockSiteInfo", DeviceConfig_GetBlockSiteInfo },
	
	{ 0, 0 }
};
/**********************************************************/

/********************* DeviceInfo Methods *********************/
static void DeviceInfo_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service: DeviceInfo:1\">\r\n"
		"      <ModelName>%s</ModelName>\r\n"
		"      <Description>Netgear Smart Wizard "SMARTAGENT_VERSION", specification "SPEC_VERSION" version</Description>\r\n"
		"      <SerialNumber>%s</SerialNumber>\r\n"
		"      <Firmwareversion>%s</Firmwareversion>\r\n"
		"      <SmartAgentversion>"SMARTAGENT_VERSION"</SmartAgentversion>\r\n"
		"      <FirewallVersion>"FIREWALL_INFO"</FirewallVersion>\r\n"
		"      <VPNVersion>"VPN_INFo"</VPNVersion>\r\n"
		"      <OthersoftwareVersion>N/A</OthersoftwareVersion>\r\n"
		"      <Hardwareversion>%s</Hardwareversion>\r\n"
		"      <Otherhardwareversion>N/A</Otherhardwareversion>\r\n"
		"      <FirstUseDate>Sunday, 30 Sep 2007 01:10:03</FirstUseDate>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[2048];
	char mode_name[64], fw_version[64], hw_version[64],seria_number[64]; 
	
	//strcpy(seria_number, cat_file(SE_NUMBER));
	strcpy(seria_number, config_get("wps_pin"));
	strcpy(mode_name, cat_file(FW_MODULE));
	strcpy(fw_version, cat_file(FW_VERSION));
	strcpy(hw_version, cat_file(HW_VERSION));
	
	sprintf(body, resp, mode_name, seria_number, fw_version, hw_version);
	BuildSendAndCloseSoapResp(req, body);
}

static void DeviceInfo_GetSysUpTime(request * req)
{
	static char resp[] =
		"   <m:GetSysUpTimeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service: DeviceInfo:1\">\r\n"
		"      <SysUpTime>%s</SysUpTime>\r\n"
		"   </m:GetSysUpTimeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], buf[64], sys_uptime[16] = "", *p;
	int total, sec, min, hour;
	FILE *f;
	
	 if ((f = fopen("/proc/uptime", "r"))) {
		fgets(buf, sizeof(buf), f);
		fclose(f);
		
		p = buf;
		while (*p != '.')	p++;
		*p = '\0';

		total = atoi(buf);
		sec = total % 60;
		total = total / 60;
		min = total % 60;
		hour = total / 60;

		if (hour < 10)
			sprintf(sys_uptime, "0%d:", hour);
		else
			sprintf(sys_uptime, "%d:", hour);
		if (min < 10)
			sprintf(sys_uptime, "%s0%d:", sys_uptime, min);
		else
			sprintf(sys_uptime, "%s%d:", sys_uptime, min);
		if (sec < 10)
			sprintf(sys_uptime, "%s0%d", sys_uptime, sec);
		else
			sprintf(sys_uptime, "%s%d", sys_uptime, sec);
	 }

	sprintf(body, resp, sys_uptime);
	BuildSendAndCloseSoapResp(req, body);
}

static void DeviceInfo_GetAttachDevice(request * req)
{
	static char resp[] =
		"   <m:GetAttachDeviceResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service: DeviceInfo:1\">\r\n"
		"      <NewAttachDevice>%s</NewAttachDevice>\r\n"
		"   </m:GetAttachDeviceResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024] = "", buf[128], info[1024] = "";
	char ip[32], mac[32], name[64];
	int num = 0;
	FILE *f;

	system("killall -SIGUSR1 net-scan");
	sleep(2);

	if ((f = fopen("/tmp/netscan/attach_device", "r")) != NULL) {
		while (fgets(buf, sizeof(buf), f)) {
			sscanf(buf, "%s %s %s %*s\n", ip, mac, name);
			sprintf(body, "%s@%d;%s;%s;%s", body, ++num, ip, name, mac);
		}
		
		if (num > 0)
			sprintf(info, "%d%s", num, body);
		fclose(f);
	}

	sprintf(body, resp, info);
	BuildSendAndCloseSoapResp(req, body);
}

static struct method_struct soapMethod2[] =
{
	{ "GetInfo",  DeviceInfo_GetInfo },
	{ "GetSysUpTime", DeviceInfo_GetSysUpTime },
	{ "GetAttachDevice", DeviceInfo_GetAttachDevice },
	
	{ 0, 0 }
};
/**********************************************************/

/******************* WANIPConnection Methods ******************/
static void update_to_file()
{
	struct portmap *item = portmap_head;
	char buff[256];
	FILE * fp;
	
	if ((fp = fopen(PORT_MAP_FILE, "w")) == NULL)
		return;

	if (item == NULL) {
		/* Clear the file to be empty */
		fclose(fp);
		return;
	}
	
	/* Format as this:  Active | Protocol | Int. Port | Ext. Port | IP Address | Desc */
	sprintf(buff, "%s;%s;%s;%s;%s;%s; @#$&*!\n",
		item->enabled,
		item->protocol,
		item->intport,
		item->extport,
		item->inthost,
		item->desc);
		fputs(buff, fp);

	fclose(fp);
}

int portmap_create()
{
	struct portmap *entry;
	char buf[256];
	FILE *fp;
	
	if ((fp = fopen(PORT_MAP_FILE, "r")) == NULL)
		return -2;
	   
	while (fgets(buf, sizeof(buf), fp)) {
		/* Format as this: Active | Protocol | Int. Port | Ext. Port | IP Address | Desc */
		char *p;
		char *active, *protocol, *intport, *extport, *dstip, *desc;
		
		active = buf;

		if ((p = strchr(buf, ';')) == NULL)
			continue;
		*p = '\0';
		protocol = p + 1;

		if ((p = strchr(protocol, ';')) == NULL)
			continue;
		*p = '\0';
		intport = p + 1;

		if ((p = strchr(intport, ';')) == NULL)
			continue;
		*p = '\0';
		extport = p + 1;

		if ((p = strchr(extport, ';')) == NULL)
			continue;
		*p = '\0';
		dstip = p + 1;

		if ((p = strchr(dstip, ';')) == NULL)
			continue;
		*p = '\0';
		desc = p + 1;

		if ((p = strchr(desc, ';')) == NULL)
			continue;
		*p = '\0';

		entry = (struct portmap*)malloc(sizeof(struct portmap));

		if (entry == NULL)
			return -1;

		if (strcasecmp(active, "YES") == 0)
			entry->enabled = 1;
		else
			entry->enabled = 0;

		strcpy(entry->protocol, protocol);
		strcpy(entry->intport, intport);
		strcpy(entry->extport, extport);
		strcpy(entry->inthost, dstip);
		sprintf(entry->desc, "%s", 
			strlen(desc) < sizeof(entry->desc) ? desc : "");

		if (portmap_tail) {
			portmap_tail->next = entry;
			entry->prev = portmap_tail;
			entry->next = NULL;
			portmap_tail = entry;
		} else {
			portmap_head = portmap_tail = entry;
			entry->prev = NULL;
			entry->next = NULL;
		}
	}

	fclose(fp);
	return 0;
}

/*
int portmap_add(char *add, char *proto, char *int_port, 
		char *ext_port, char *int_ip, char *desc)
{
	struct portmap *item = portmap_head;
	   
	while (item != NULL) {
		if (strcasecmp(item->protocol, proto) == 0
		    && strcmp(item->extport, ext_port) == 0) {
			if (strcmp(item->inthost, int_ip) == 0)
				return -3;

			strcpy(item->intport, int_port);
			sprintf(item->desc, "%s", 
				strlen(desc) < sizeof(item->desc) ? desc : "");

			return 0;
		}
		
		item = item->next;
	}

	if (item == NULL) {
		struct portmap *entry = (struct portmap*)malloc(sizeof(struct portmap));

		if (entry == NULL)
			return -1;

		entry->enabled = 1;
		strcpy(entry->protocol, proto);
		strcpy(entry->intport, int_port);
		strcpy(entry->extport, ext_port);
		strcpy(entry->inthost, int_ip);
		sprintf(entry->desc, "%s", 
			strlen(desc) < sizeof(entry->desc) ? desc : "");

		if (portmap_tail) {
			portmap_tail->next = entry;
			entry->prev = portmap_tail;
			entry->next = NULL;
			portmap_tail = entry;
		} else {
			portmap_head = portmap_tail = entry;
			entry->prev = NULL;
			entry->next = NULL;
		}
	}

	return 0;
}

int portmap_del(char *proto, char *ext_port)
{
	struct portmap *item = portmap_head;
	   
	while (item != NULL) {
		if (strcasecmp(item->protocol, proto) == 0
		    && strcmp(item->extport, ext_port) == 0) {
			if (item == portmap_head) {
				if (item->next == NULL)
					portmap_head = portmap_tail = NULL;
				else {
					portmap_head = item->next;
					portmap_head->prev = NULL;
				}
			} else if (item == portmap_tail) {
				portmap_tail = portmap_tail->prev;
				portmap_tail->next = NULL;
			} else {
				item->prev->next = item->next;
				item->next->prev = item->prev;
			}

			item->next = item->prev = NULL;
			free(item);
			return 0;
		}
		
		item = item->next;
	}

	return -1;
}
*/

void free_portmap_queue(void)
{
	struct portmap *tmp, *next;

	tmp = portmap_head;

	while (tmp) {
		next = tmp->next;
		tmp->next =tmp->prev = NULL;
		free(tmp);
		tmp = next;
	}

	portmap_head = portmap_tail = NULL;
}

int get_portmap_items(char *info)
{
	struct portmap *item = portmap_head;
	int num = 0;

	while (item) {
		if (num++ > 0)
			sprintf(info, "%s@", info);
		
		sprintf(info, "%s%s;%s;%s;%s;%s; @#$&*!", 
			info,
			item->protocol,
			item->intport,
			item->extport,
			item->inthost,
			item->desc);

		item = item->next;
	}

	return num;
}

void get_conntype(char *type)
{
	sprintf(type, config_get("wan_proto"));

	if (strcmp(type, "dhcp") == 0)
		strcpy(type, "DHCP");
	else if (strcmp(type, "static") == 0)
		strcpy(type, "Static");
	else if (strcmp(type, "pppoe") == 0)
		strcpy(type, "PPPoE");
	else if (strcmp(type, "pptp") == 0)
		strcpy(type, "PPTP");
	else if (strcmp(type, "bigpond") == 0)
		strcpy(type, "BigPond");
	else
		strcpy(type, "Unknown");
}

void change_mac_format(char *src, char *dst)
{
	int i=0, j=0;
	while(src[i] != '\0' && src[i] != '\n') {
		if (src[i] != ':')
			dst[j++] = src[i];
		i++;
	}
	dst[j] = '\0';
}

static void WANIPConnection_SetConnectionType(request * req)
{
	static char resp[] =
		"   <m:SetConnectionTypeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"      <NewConnectionType>%s</NewConnectionType>\r\n"
		"   </m:SetConnectionTypeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *type, *loginname, *password, *idle, *conn_mode;
	char *pptp_ip, *pptp_serip, *conn_id, *bpa_server, *pri_dns, *sec_dns;
	int i;

	char body[1024], wan_status[32] = "Down", idle_time[16], wan_endis_dod[8] = "1";
	FILE *fp;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		type = GetValueFromNameValueList(&data, "NewConnectionType");
		pri_dns = GetValueFromNameValueList(&data, "NewPrimaryDNS");
		sec_dns = GetValueFromNameValueList(&data, "NewSecondaryDNS");

		if (type && ((strcasecmp(type, "Static") == 0) || (strcasecmp(type, "StaticIP") == 0))) {
			config_set("wan_proto", "static");
			config_set("internet_type", "1");
			config_set("wan_ether_wan_assign", "1");

			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0) {
				config_set("wan_ether_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_ether_dns_assign", "0");
			goto end;
		} else if (type && (strcasecmp(type, "DHCP") == 0) ) {
			config_set("wan_proto", "dhcp");
			config_set("internet_type", "1");
			config_set("wan_ether_wan_assign", "0");

			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0) {
				config_set("wan_ether_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_ether_dns_assign", "0");
			
			goto end;
		} else if (type && strcasecmp(type, "PPPoE") == 0) {
			loginname = GetValueFromNameValueList(&data, "NewISPLoginname");
			password = GetValueFromNameValueList(&data, "NewISPPassword");
			idle = GetValueFromNameValueList(&data, "Newidletimer");
			conn_mode  = GetValueFromNameValueList(&data, "NewConnectionMode");
			
			config_set("wan_proto", "pppoe");
			config_set("internet_type", "0");
			config_set("internet_ppp_type", "0");
			config_set("wan_pppoe_wan_assign", "0");
			config_set("wan_pppoe_username", loginname);
			config_set("wan_pppoe_passwd", password);

			if (strcasecmp(conn_mode, "AlwaysOn") == 0)
				sprintf(wan_endis_dod, "0");
			else if (strcasecmp(conn_mode, "ManuallyConnect") == 0)
				sprintf(wan_endis_dod, "2");
			else
				sprintf(wan_endis_dod, "1");
				
			config_set("wan_pppoe_demand", wan_endis_dod);
			config_set("wan_endis_dod", wan_endis_dod);

			if(idle){
				i = atoi(idle); sprintf(idle_time, "%d", i * 60); // min --> sec
				config_set("wan_pppoe_idletime", idle_time);
			}
			
			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0) {
				config_set("wan_pppoe_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_pppoe_dns_assign", "0");
				
			goto end;
		} else if (type && strcasecmp(type, "PPTP") == 0) {
			loginname = GetValueFromNameValueList(&data, "NewISPLoginname");
			password = GetValueFromNameValueList(&data, "NewISPPassword");
			idle = GetValueFromNameValueList(&data, "Newidletimer");
			conn_mode  = GetValueFromNameValueList(&data, "NewConnectionMode");
			pptp_ip = GetValueFromNameValueList(&data, "NewPPTPMyIP");
			pptp_serip = GetValueFromNameValueList(&data, "NewPPTPServerIP");
			conn_id  = GetValueFromNameValueList(&data, "NewConnectionID");

			config_set("wan_proto", "pptp");
			config_set("internet_type", "0");
			config_set("internet_ppp_type", "1");
			config_set("wan_pptp_username", loginname);
			config_set("wan_pptp_password", password);
			config_set("wan_pptp_local_ip", pptp_ip);
			config_set("wan_pptp_server_ip", pptp_serip);
			config_set("wan_pptp_connection_id", conn_id);

			if (pptp_ip /*&& *pptp_ip != '\0'*/)
				config_set("wan_pptp_wan_assign", "1");
			else
				config_set("wan_pptp_wan_assign", "0");

			if (strcasecmp(conn_mode, "AlwaysOn") == 0)
				sprintf(wan_endis_dod, "0");	
			else if (strcasecmp(conn_mode, "ManuallyConnect") == 0)
				sprintf(wan_endis_dod, "2");
			else
				sprintf(wan_endis_dod, "1");
				
			config_set("wan_pptp_demand", wan_endis_dod);
			config_set("wan_endis_dod", wan_endis_dod);

			if(idle){
				i = atoi(idle); sprintf(idle_time, "%d", i * 60); // min --> sec
				config_set("wan_pptp_idle_time", idle_time);
			}

			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0) {
				config_set("wan_pptp_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_pptp_dns_assign", "0");
			
			goto end;
		} else if (type && ((strcasecmp(type, "BigPond") == 0) || (strcasecmp(type, "TelstaBigP") == 0))) {
			/*loginname = GetValueFromNameValueList(&data, "NewISPLoginname");
			password = GetValueFromNameValueList(&data, "NewISPPassword");
			idle = GetValueFromNameValueList(&data, "Newidletimer");
			conn_mode  = GetValueFromNameValueList(&data, "NewConnectionMode");
			bpa_server = GetValueFromNameValueList(&data, "NewBigpondAuthServer");

			config_set("wan_proto", "bigpond");
			config_set("internet_type", "0");
			config_set("internet_ppp_type", "2");
			config_set("wan_bpa_username", loginname);
			config_set("wan_bpa_password", password);
			config_set("wan_bpa_servicename", bpa_server);

			if (strcasecmp(conn_mode, "AlwaysOn") == 0)
				sprintf(wan_endis_dod, "0");
			else if (strcasecmp(conn_mode, "ManuallyConnect") == 0)
				sprintf(wan_endis_dod, "2");
			else
				sprintf(wan_endis_dod, "1");
			config_set("wan_bpa_demand", wan_endis_dod);
			config_set("wan_endis_dod", wan_endis_dod);

			i = atoi(idle); sprintf(idle_time, "%d", i * 60); // min --> sec
			config_set("wan_bpa_idle_time", idle_time);

			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0) {
				config_set("wan_bpa_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_bpa_dns_assign", "0");
			*/
			SoapError(req, "402");
			//goto end;
			return;
		} else{
			SoapError(req, "402");
		}

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");

	return;

end:
	if ((fp = fopen("/tmp/WAN_status", "r"))) {
		fgets(wan_status, sizeof(wan_status), fp);
		fclose(fp);
		}
	if (type && strcmp(wan_status, "Link down") != 0)
		strcpy(wan_status, type);

	//system("/usr/sbin/nvram commit");
	do_cmd("/sbin/cmdwan.sh restart");
	
	sprintf(body, resp, wan_status);
	BuildSendAndCloseSoapResp(req, body);
}

int check_samesubnet(char *lan_ip, char *lan_subnet, char *wan_ip, char *wan_subnet)
{
	u32 ip_a, ip_b, subnet_a, subnet_b;

	ip_a = inet_addr(lan_ip);
	ip_b = inet_addr(wan_ip);

	subnet_a = inet_addr(lan_subnet);
	subnet_b = inet_addr(wan_subnet);

	return (((ip_a & subnet_a) == (ip_b & subnet_a)) || ((ip_a & subnet_b) == (ip_b & subnet_b)));
}

static void WANIPConnection_SetIPInterfaceInfo(request * req)
{
	static char resp[] =
		"   <m:SetIPInterfaceInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"   </m:SetIPInterfaceInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *type, *ipaddr, *netmask, *gateway, *pri_dns, *sec_dns;
	char lan_ip[32], lan_subnet[32];

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		type = GetValueFromNameValueList(&data, "NewAddressingType");
		ipaddr = GetValueFromNameValueList(&data, "NewExternalIPAddress");
		netmask = GetValueFromNameValueList(&data, "NewSubnetMask");
		gateway = GetValueFromNameValueList(&data, "NewDefaultGateway");
		pri_dns = GetValueFromNameValueList(&data, "NewPrimaryDNS");
		sec_dns = GetValueFromNameValueList(&data, "NewSecondaryDNS");

		if (type && ipaddr && netmask && gateway && pri_dns && ((strcasecmp(type, "Static") == 0) || (strcasecmp(type, "StaticIP") == 0))
		    && (strcasecmp(ipaddr, "...") != 0) && (strcasecmp(netmask, "...") != 0) 
		    && (strcasecmp(gateway, "...") != 0) && (strcasecmp(pri_dns, "...") != 0)) {
			config_set("wan_proto", "static");
			config_set("internet_type", "1");
			config_set("wan_ether_wan_assign", "1");
			config_set("wan_ipaddr", ipaddr);
			config_set("wan_netmask", netmask);
			config_set("wan_gateway", gateway);

			if (pri_dns && strcmp(pri_dns, "0.0.0.0") != 0 && strcmp(pri_dns, "...") != 0) {
				config_set("wan_ether_dns_assign", "1");
				config_set("wan_ether_dns1", pri_dns);
				if (sec_dns && strcmp(sec_dns, "...") != 0)
					config_set("wan_ether_dns2", sec_dns);
				else
					config_set("wan_ether_dns2", "0.0.0.0");
			} else
				config_set("wan_ether_dns_assign", "0");

			sprintf(lan_ip, config_get("lan_ipaddr"));
			sprintf(lan_subnet, config_get("lan_netmask"));
			 if (check_samesubnet(lan_ip, lan_subnet, ipaddr, netmask)) {
				do_cmd("/sbin/cmdscflict.sh");
				do_cmd("/sbin/cmdwan.sh restart");
				//do_two_cmd("/sbin/cmdscflict", "/sbin/ip_conflict.sh start");
			 	//do_two_cmd("/sbin/cmdlan restart", "/sbin/cmdwan restart");
			 } else
			 	do_cmd("/sbin/cmdwan.sh restart");

			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANIPConnection_SetMACAddress(request * req)
{
	static char resp[] =
		"   <m:SetMACAddressResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"   </m:SetMACAddressResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *type, *mac, wan_proto[32], mac_addr[32];
	int i =0;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		type = GetValueFromNameValueList(&data, "NewMACAddressOverride");
		mac = GetValueFromNameValueList(&data, "NewMACAddress");
		sprintf(wan_proto, config_get("wan_proto"));

		if (type && strcmp(type, "2")==0 ) {
	              if( strlen(mac) != 12)
				  	SoapError(req, "402");

			for(i=0;i<17;i++)
                     {
                                 if(((i+1)%3) == 0)
	       			       mac_addr[i]=':';
				     else
					 	 mac_addr[i]=mac[2*(i/3)+(i%3)];

			}
			mac_addr[17]='\0';

			if (strcmp(wan_proto, "static") == 0 || strcmp(wan_proto, "dhcp") == 0) {
				config_set("wan_ether_mac_assign", type);
				config_set("wan_ether_this_mac", mac_addr);
			} else if (strcmp(wan_proto, "pptp") == 0 ) {
				config_set("wan_pptp_mac_assign", type);
				config_set("wan_pptp_this_mac", mac_addr);
			} else if (strcmp(wan_proto, "bigpond") == 0 ) {
				config_set("wan_bpa_mac_assign", type);
				config_set("wan_bpa_this_mac", mac_addr);
			}
			
                     do_cmd("/usr/sbin/nvram commit");
			do_cmd("/sbin/cmdwan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else if (type) {

		       if(*type == '0')
			   	*type='1';
			else  if(*type == '1')
			   	*type='0';
			if (strcmp(wan_proto, "static") == 0 || strcmp(wan_proto, "dhcp") == 0)
				config_set("wan_ether_mac_assign", type);
			else if (strcmp(wan_proto, "pptp") == 0 )
				config_set("wan_pptp_mac_assign", type);
			else if (strcmp(wan_proto, "bigpond") == 0 )
				config_set("wan_bpa_mac_assign", type);
			
                    // do_cmd("/usr/sbin/nvram commit");
			do_cmd("/sbin/cmdwan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANIPConnection_SetSmartWizardDetection(request * req)
{
	static char resp[] =
		"   <m:SetSmartWizardDetectionResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"      <NewConnectionType>%s</NewConnectionType>\r\n"
		"   </m:SetSmartWizardDetectionResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], status[16] = "10000", type[16] = "Down", *wan_status;
	FILE *fp;

	if (config_state == CONFIG_STARTED) {
		wan_status = cat_file("/tmp/WAN_status");
		if (*wan_status == '\0' || strcmp(wan_status, "Link down") == 0) {
			sprintf(body, resp, "Down");
			BuildSendAndCloseSoapResp(req, body);
			return;
		}

		if (!Internet_Detection(type)) {
			SoapError(req, "501");
			return;
		}

		sprintf(body, resp, type);
		BuildSendAndCloseSoapResp(req, body);
		
		strcpy(detect_conntype, type);
		detect_state = CONFIG_STARTED;
	} else
		SoapError(req, "501");
}

static void WANIPConnection_SetMaxMTUSize(request * req)
{
	static char resp[] =
		"   <m:SetMaxMTUSizeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"   </m:SetMaxMTUSizeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *mtu, wan_proto[32];

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		mtu = GetValueFromNameValueList(&data, "NewMaxMTUSize");
		sprintf(wan_proto, config_get("wan_proto"));

		if (mtu) {
			if (strcmp(wan_proto, "pptp") == 0 )
				config_set("wan_pptp_mtu", mtu);
			else if (strcmp(wan_proto, "pppoe") == 0 )
				config_set("wan_pppoe_mtu", mtu);
			else
				config_set("wan_dhcp_mtu", mtu);
			
			do_cmd("/sbin/cmdwan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANIPConnection_AddPortMapping(request * req)
{
	static char resp[] =
		"   <m:AddPortMappingResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"   </m:AddPortMappingResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *protocol, *int_port, *ext_port, *int_ip, *desc, *enable_add;
	struct portmap *entry;
	int ret = 0;
	char buff[256];
	FILE *fp;
	char *pp;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		protocol = GetValueFromNameValueList(&data, "NewProtocol");
		int_port = GetValueFromNameValueList(&data, "NewInternalPort");
		ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
		int_ip = GetValueFromNameValueList(&data, "NewInternalClient");
		desc = GetValueFromNameValueList(&data, "NewPortMappingDescription");
		enable_add = GetValueFromNameValueList(&data, "NewPortMappingAdd");

		if (protocol && int_port && ext_port && int_ip && enable_add) {
				entry->enabled=1;
				strcpy(entry->protocol, protocol);
				strcpy(entry->intport, int_port);
				strcpy(entry->extport, ext_port);
				strcpy(entry->inthost, int_ip);
				sprintf(entry->desc, "%s", strlen(desc) < sizeof(entry->desc) ? desc : "");
				//update_to_file();
				fp = fopen(SOAP_PMLIST_FILE, "w");
	
				/* Format as this:  Active | Protocol | Int. Port | Ext. Port | IP Address | Desc */
				sprintf(buff, "%d;%s;%s;%s;%s;%s; @#$&*!\n",
					entry->enabled,
					entry->protocol,
					entry->intport,
					entry->extport,
					entry->inthost,
					entry->desc);
					fputs(buff, fp);
				fclose(fp);

				//signal miniigd task to do addPortmaping
				//atoi(cat_file("/var/run/miniupnp.pid"))
				system("killall -SIGUSR1 miniigd");

				//sleep a while, then check if success
		#if 0
				sleep(1);
				pp = cat_file(SOAP_PMLIST_FILE);
				if( (strlen(pp) == 1) && (atoi(pp) == 1) )
				{
					fp = fopen(SOAP_PMLIST_FILE, "w");
					fputs("success", fp);
					BuildSendAndCloseSoapResp(req, resp);
				}
				else
				{
					char tmp[24];
					fp = fopen(SOAP_PMLIST_FILE, "w");
					sprintf(tmp,"failed, len[%d]\n", strlen(pp));
					//fputs("failed\n", fp);
					fputs(tmp, fp);
					SoapError(req, "501");  //action failed
				}
				fclose(fp);
				//unlink(SOAP_PMLIST_FILE);
				//BuildSendAndCloseSoapResp(req, resp);
		#else
				sleep(1);
				pp = cat_file(SOAP_RESULT_FILE);
				if( (strlen(pp) == 1) && (atoi(pp) == 1) )
				{
					//fp = fopen(SOAP_PMLIST_FILE, "w");
					//fputs("success", fp);
					BuildSendAndCloseSoapResp(req, resp);
				}
				else
				{
					//char tmp[24];
					//fp = fopen(SOAP_PMLIST_FILE, "w");
					//sprintf(tmp,"failed, len[%d]\n", strlen(pp));
					//fputs("failed\n", fp);
					//fputs(tmp, fp);
					SoapError(req, "501");  //action failed
				}
				unlink(SOAP_PMLIST_FILE);
				unlink(SOAP_RESULT_FILE);
		#endif
				
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANIPConnection_DeletePortMapping(request * req)
{
	static char resp[] =
		"   <m:DeletePortMappingResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"   </m:DeletePortMappingResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *protocol, *ext_port, *enable_add;
	struct portmap *entry;
	int ret = 0;
	char buff[256];
	FILE *fp;
	char *pp;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		protocol = GetValueFromNameValueList(&data, "NewProtocol");
		ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
		enable_add = GetValueFromNameValueList(&data, "NewPortMappingAdd");

		if (protocol && ext_port && enable_add) {
				entry->enabled=0;
				strcpy(entry->protocol, protocol);
				strcpy(entry->intport, " ");
				strcpy(entry->extport, ext_port);
				strcpy(entry->inthost, " ");
				strcpy(entry->desc, " ");
				//update_to_file(entry);
				fp = fopen(SOAP_PMLIST_FILE, "w");
	
				/* Format as this:  Active | Protocol | Int. Port | Ext. Port | IP Address | Desc */
				sprintf(buff, "%d;%s;%s;%s;%s;%s; @#$&*!\n",
					entry->enabled,
					entry->protocol,
					entry->intport,
					entry->extport,
					entry->inthost,
					entry->desc);
					fputs(buff, fp);
				fclose(fp);

				//signal miniigd task to do addPortmaping
				//atoi(cat_file("/var/run/miniupnp.pid"))
				system("killall -SIGUSR1 miniigd");

				//sleep a while, then check if success
#if 0				
				sleep(1);
				pp = cat_file(SOAP_PMLIST_FILE);
				if( (strlen(pp) == 1) && (atoi(pp) == 1) )
				{
					fp = fopen(SOAP_PMLIST_FILE, "w");
					fputs("success", fp);
					BuildSendAndCloseSoapResp(req, resp);
				}
				else
				{
					char tmp[64];
					fp = fopen(SOAP_PMLIST_FILE, "w");
					sprintf(tmp,"failed, len[%d]\n", strlen(pp));
					//fputs("failed\n", fp);
					fputs(tmp, fp);
					SoapError(req, "501");  //action failed
				}
				fclose(fp);
				//unlink(SOAP_PMLIST_FILE);
#else
				sleep(1);
				pp = cat_file(SOAP_RESULT_FILE);
				if( (strlen(pp) == 1) && (atoi(pp) == 1) )
					BuildSendAndCloseSoapResp(req, resp);
				else
					SoapError(req, "501");  //action failed

				unlink(SOAP_PMLIST_FILE);
				unlink(SOAP_RESULT_FILE);
#endif			
				//do_cmd("/usr/sbin/net-wall start");
				//BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANIPConnection_GetConnectionTypeInfo(request * req)
{
	static char resp[] =
		"   <m:GetConnectionTypeInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"      <NewConnectionType>%s</NewConnectionType>\r\n"
		"   </m:GetConnectionTypeInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], type[16] = "Down";
	char *wan_status;

	if (detect_state == CONFIG_STARTED) {
		strcpy(type, detect_conntype);
		detect_state = CONFIG_FINISHED;
	} else
		get_conntype(type);
	
	wan_status = cat_file("/tmp/WAN_status");
	if (*wan_status == '\0' || strcmp(wan_status, "Link down") == 0)
		strcpy(type, "Down");
	
	sprintf(body, resp, type);
	BuildSendAndCloseSoapResp(req, body);
}

static void WANIPConnection_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"      <NewEnable>%s</NewEnable>\r\n"
		"      <NewConnectionType>%s</NewConnectionType>\r\n"
		"      <NewExternalIPAddress>%s</NewExternalIPAddress>\r\n"
		"      <NewSubnetMask>%s</NewSubnetMask>\r\n"
		"      <NewAddressingType>%s</NewAddressingType>\r\n"
		"      <NewDefaultGateway>%s</NewDefaultGateway>\r\n"
		"      <NewMACAddress>%s</NewMACAddress>\r\n"
		"      <NewMACAddressOverride>%s</NewMACAddressOverride>\r\n"
		"      <NewMaxMTUSize>%s</NewMaxMTUSize>\r\n"
		"      <NewDNSEnabled>%s</NewDNSEnabled>\r\n"
		"      <NewDNSServers>%s</NewDNSServers>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";
	
	char body[2048], enable[8];
	char mac_addr[32] = "";
	char override[8], mtusize[32],  dns_enable[8], dns_server[128] = "";

	char wan_proto[16] = "Down", wan_assign[8], dns_assign[8];
	char wan_ip[32] = "0.0.0.0", wan_mask[32] = "0.0.0.0";
	char wan_gateway[32] = "0.0.0.0";
	char status[8] = "", *wan_status;
	char buf[32];
	char line[64];
	char *q;
	FILE *f;

	sprintf(enable, "1");
	sprintf(override, "0");
	sprintf(dns_enable, "1");
	
	sprintf(wan_proto, config_get("wan_proto"));	
	if (strcmp(wan_proto, "dhcp") == 0) {
		sprintf(wan_ip, config_get("wan_dhcp_ipaddr"));
		sprintf(wan_mask, config_get("wan_dhcp_netmask"));
		sprintf(wan_gateway, config_get("wan_dhcp_gateway"));
		sprintf(mtusize, config_get("wan_dhcp_mtu"));
		strcpy(buf, config_get("wan_hwaddr"));
		change_mac_format(buf, mac_addr);

		if (wan_ip[0] == '\0' || strcmp(wan_ip, "0.0.0.0") == 0) {
			sprintf(wan_gateway, "0.0.0.0");
			sprintf(wan_mask, "0.0.0.0");
		}
		
		sprintf(wan_proto, "DHCP");
	} else if (strcmp(wan_proto, "static") == 0) {
		sprintf(wan_ip, config_get("wan_ipaddr"));
		sprintf(wan_gateway, config_get("wan_gateway"));
		sprintf(wan_mask, config_get("wan_netmask"));
		sprintf(mtusize, config_get("wan_dhcp_mtu"));
		strcpy(buf, config_get("wan_hwaddr"));
		change_mac_format(buf, mac_addr);

		sprintf(wan_proto, "Static");
	/*} else if (strcmp(wan_proto, "bigpond") == 0) {
		sprintf(wan_ip, config_get("wan_dhcp_ipaddr"));
		sprintf(wan_gateway, config_get("wan_dhcp_gateway"));
		sprintf(wan_mask, config_get("wan_dhcp_netmask"));
		sprintf(mtusize, config_get("wan_dhcp_mtu"));
		strcpy(buf, config_get("wan_hwaddr"));
		change_mac_format(buf, mac_addr);

		sprintf(wan_proto, "BigPond");*/
	} else if (strcmp(wan_proto, "pptp") == 0) {
		sprintf(wan_assign, config_get("wan_pptp_wan_assign"));
		sprintf(mtusize, config_get("wan_pptp_mtu"));
		strcpy(buf, config_get("wan_hwaddr"));
		change_mac_format(buf, mac_addr);

		if ((f = fopen("/tmp/ppp/ppp0-status", "r"))) {
			fgets(status, sizeof(status), f);
			fclose(f);
		}

		if (status && status[0] == '1') {
			//q = get_if_addr("ppp0");
			q=config_get("wan0_ipaddr");
			if (q) {
				sprintf(wan_ip, q);
				sprintf(wan_gateway, q);
				sprintf(wan_mask, "255.255.255.255");
			} else if (wan_assign && wan_assign[0] == '1') {
				sprintf(wan_ip, config_get("wan_pptp_local_ip"));
				sprintf(wan_gateway, "0.0.0.0");
				sprintf(wan_mask, "0.0.0.0");
			}
		}
		if (wan_ip[0] == '\0') {
			sprintf(wan_ip, "0.0.0.0");
			sprintf(wan_mask, "0.0.0.0");
			sprintf(wan_gateway, "0.0.0.0");
		}
		
		sprintf(wan_proto, "PPTP");
	} else if (strcmp(wan_proto, "pppoe") == 0) {
		sprintf(wan_assign, config_get("wan_pppoe_wan_assign"));
		sprintf(mtusize, config_get("wan_pppoe_mtu"));
		strcpy(buf, config_get("wan_hwaddr"));
		change_mac_format(buf, mac_addr);

		if ((f = fopen("/tmp/ppp/ppp0-status", "r"))) {
			fgets(status, sizeof(status), f);
			fclose(f);
		}

		if (status && status[0] == '1') {
			//q = get_if_addr("ppp0");
			q=config_get("wan0_ipaddr");
			if (q) {
				sprintf(wan_ip, q);
				sprintf(wan_gateway, q);
				sprintf(wan_mask, "255.255.255.255");
			} else if (wan_assign && wan_assign[0] == '1') {
				sprintf(wan_ip, config_get("wan_pppoe_ip"));
				sprintf(wan_gateway, "0.0.0.0");
				sprintf(wan_mask, "0.0.0.0");
			}
		}
		if (wan_ip[0] == '\0') {
			sprintf(wan_ip, "0.0.0.0");
			sprintf(wan_mask, "0.0.0.0");
			sprintf(wan_gateway, "0.0.0.0");
		}
		
		sprintf(wan_proto, "PPPoE");
	}

	if ((f = fopen("/tmp/resolv.conf", "r"))) {
		while (fgets(line, sizeof(line), f)) {
			char *token = strtok(line, " \t\n");
			if (!token || strcmp(token, "nameserver") != 0)
				continue;
			if (!(token = strtok(NULL, " \t\n")))
				continue;
				
			sprintf(dns_server, "%s%s ", dns_server, token);
		}	
		fclose(f);
	}

	if (wan_ip[0] == '\0')
		sprintf(wan_ip, "0.0.0.0");

	if (dns_server[0] == '\0')
		sprintf(dns_server, "0.0.0.0");

	wan_status = cat_file("/tmp/WAN_status");
	if (*wan_status == '\0' || strcmp(wan_status, "Link down") == 0) {
		sprintf(wan_proto, "Down");
		sprintf(wan_ip, "0.0.0.0");
		sprintf(wan_mask, "0.0.0.0");
		sprintf(wan_gateway, "0.0.0.0");
		sprintf(dns_server, "0.0.0.0");
	}
	
	sprintf(body, resp, enable, wan_proto, wan_ip, wan_mask, wan_proto, 
		wan_gateway, mac_addr, override, mtusize, dns_enable, dns_server);

	BuildSendAndCloseSoapResp(req, body);
}

static void WANIPConnection_GetPortMappingInfo(request * req)
{
	static char resp[] =
		"   <m:GetPortMappingInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANIPConnection:1\">\r\n"
		"      <NewPortMappingNumberOfEntries>%d</NewPortMappingNumberOfEntries>\r\n"
		"      <NewPortMappingInfo>%s</NewPortMappingInfo>\r\n"
		"   </m:GetPortMappingInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[2048], info[1024] = "";
	int ret = 0, num = 0;

	struct portmap *entry;
	char buf[256], tmp[256];
	char *p;
	char *token=NULL; 
	char *int_ip, *int_port, *ext_port, *protocol, *desc; 
	FILE *fp;
	
	
	if((fp = fopen(PORT_MAP_FILE, "r"))){
		while (fgets(tmp, sizeof(tmp), fp)) {
			p=tmp;
			strsep(&p,";");
			protocol=strsep(&p,";");
			int_port=strsep(&p,";");
			ext_port=strsep(&p,";");
			int_ip=strsep(&p,";");
			desc=strsep(&p,";");
			sprintf(buf,"%s;%s;%s;%s;%s",protocol, int_port, ext_port, int_ip, desc);
			sprintf(buf, "%s@", buf);
			strcat(info, buf);
			num++;
		}
		fclose(fp);
	}
	sprintf(body, resp, num, info);
	BuildSendAndCloseSoapResp(req, body);
	
}

static struct method_struct soapMethod3[] =
{
	{ "SetConnectionType",  WANIPConnection_SetConnectionType },
	{ "SetIPInterfaceInfo", WANIPConnection_SetIPInterfaceInfo },
	{ "SetMACAddress", WANIPConnection_SetMACAddress },
	{ "SetSmartWizardDetection",  WANIPConnection_SetSmartWizardDetection },
	{ "SetMaxMTUSize", WANIPConnection_SetMaxMTUSize },
	{ "AddPortMapping", WANIPConnection_AddPortMapping },
	{ "DeletePortMapping",  WANIPConnection_DeletePortMapping },
	
	{ "GetConnectionTypeInfo", WANIPConnection_GetConnectionTypeInfo },
	{ "GetInfo", WANIPConnection_GetInfo },
	{ "GetPortMappingInfo", WANIPConnection_GetPortMappingInfo },
	
	{ 0, 0 }
};
/**********************************************************/

/**************** WANEthernetLinkConfig Methods *****************/
static void WANEthernetLinkConfig_GetEthernetLinkStatus(request * req)
{
	static char resp[] =
		"   <m:GetEthernetLinkStatusResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANEthernetLinkConfig:1\">\r\n"
		"      <NewEthernetLinkStatus>%s</NewEthernetLinkStatus>\r\n"
		"   </m:GetEthernetLinkStatusResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], status[16] = "Down", wan_status[32] = "Link down", *p = NULL;
	FILE *fp;

	if ((fp = fopen("/tmp/WAN_status", "r"))) {
		fgets(wan_status, sizeof(wan_status), fp);
		fclose(fp);

		if (strcmp(wan_status, "Link down") == 0)
			strcpy(status, "Down");
		else
			strcpy(status, "Up");
	}

	sprintf(body, resp, status);
	BuildSendAndCloseSoapResp(req, body);
}

static void WANEthernetLinkConfig_SetWANRelease(request * req)
{
	static char resp[] =
		"   <m:SetWANReleaseResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANEthernetLinkConfig:1\">\r\n"
		"   </m:SetWANReleaseResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable, wan_proto[16];

	sprintf(wan_proto, config_get("wan_proto"));

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewWANRelease");

		if (enable && (*enable == '1')) {
			if (strcmp(wan_proto, "dhcp") == 0)
				system("/usr/bin/killall -SIGUSR2 udhcpc");
			else
				system("/sbin/cmdwan.sh stop");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");
		
		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WANEthernetLinkConfig_SetWANRenew(request * req)
{
	static char resp[] =
		"   <m:SetWANRenewResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WANEthernetLinkConfig:1\">\r\n"
		"   </m:SetWANRenewResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable, wan_proto[16];

	sprintf(wan_proto, config_get("wan_proto"));

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewWANRenew");

		if (enable && (*enable == '1')) {
			if (strcmp(wan_proto, "dhcp") == 0)
				system("/usr/bin/killall -SIGUSR1 udhcpc");
			else
				system("/sbin/cmdwan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static struct method_struct soapMethod4[] =
{
	{ "GetEthernetLinkStatus",  WANEthernetLinkConfig_GetEthernetLinkStatus },
	
	{ "SetWANRelease", WANEthernetLinkConfig_SetWANRelease },
	{ "SetWANRenew", WANEthernetLinkConfig_SetWANRenew },
	
	{ 0, 0 }
};
/**********************************************************/

/****************** LANConfigSecurity Methods *******************/

int bad_mask(char *mask)
{
	uint32_t addr;
	
	if (inet_aton(mask, (struct in_addr *)&addr) == 0)
		return 1;

	addr = ~addr;
	addr = ntohl(addr);
	
	if (addr & (addr+1))
		return 1;
	
	return 0;
}

static void LANConfigSecurity_SetConfigLANSubnet(request * req)
{
	static char resp[] =
		"   <m:SetConfigLANSubnetResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:LANConfigSecurity:1\">\r\n"
		"   </m:SetConfigLANSubnetResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *lan_subnet;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		lan_subnet = GetValueFromNameValueList(&data, "NewLANSubnet");

		if ((strncmp(lan_subnet, "255", 3) == 0) 
		    && (strcmp(lan_subnet, "255.255.255.255") != 0) 
		    && (bad_mask(lan_subnet) == 0)) {
		    	config_set("lan_netmask", lan_subnet);
			system("/sbin/cmdlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "702");
		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void LANConfigSecurity_SetConfigPassword(request * req)
{
	static char resp[] =
		"   <m:SetConfigPasswordResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:LANConfigSecurity:1\">\r\n"
		"   </m:SetConfigPasswordResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char pwd_len[16];
	char *pwd;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		pwd = GetValueFromNameValueList(&data, "NewPassword");

		if (pwd) {
			config_set("password1", pwd);
			//sprintf(pwd_len, "%d", strlen(pwd));
			//config_set("router_passwd_length", pwd_len);
		}
		system("/usr/sbin/nvram commit");
		BuildSendAndCloseSoapResp(req, resp);
		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void LANConfigSecurity_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:LANConfigSecurity:1\">\r\n"
		"      <NewLANSubnet>%s</NewLANSubnet>\r\n"
		"      <NewPassword>%s</NewPassword>\r\n"
		"      <NewWANLAN_Subnet_Match>%s</NewWANLAN_Subnet_Match>\r\n"
		"      <NewLANMACAddress>%s</NewLANMACAddress>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[2048], buf[32];
	char lan_subnet[64], password[64], match[8] = "0", lan_mac[32], status[8];
	char wan_proto[16] = "", wan_ip[]= "0.0.0.0", wan_subnet[64] = "";
	char *wan_status, *q;
	int i = 0, j = 0;
	FILE *f;

	sprintf(lan_subnet, config_get("lan_netmask"));
	sprintf(password, config_get("password1"));
	
	sprintf(buf, config_get("lan_hwaddr"));
	while(buf[i] != '\0' && buf[i] != '\n') {
		if (buf[i] != ':')
			lan_mac[j++] = buf[i];
		i++;
	}
	lan_mac[j] = '\0';

	wan_status = cat_file("/tmp/WAN_status");
	if (*wan_status == '\0' || strcmp(wan_status, "Link down") == 0)
		sprintf(wan_subnet, "0.0.0.0");
	else {
		sprintf(wan_proto, config_get("wan_proto"));	
		if (strcmp(wan_proto, "dhcp") == 0 || strcmp(wan_proto, "bigpond") == 0) {
			sprintf(wan_ip, config_get("wan_dhcp_ipaddr"));
			sprintf(wan_subnet, config_get("wan_dhcp_netmask"));
		} else if (strcmp(wan_proto, "static") == 0) {
			sprintf(wan_ip, config_get("wan_ipaddr"));
			sprintf(wan_subnet, config_get("wan_netmask"));
		} else {
			if ((f = fopen("/etc/ppp/ppp0-status", "r"))) {
				fgets(status, sizeof(status), f);
				fclose(f);
			}

			if (status && status[0] == '1') {
				q = get_if_addr("ppp0");
				if (q) {
					sprintf(wan_ip, q);
					sprintf(wan_subnet, "255.255.255.255");
				}
			}
		}

		if (wan_ip[0] == '\0' || strcmp(wan_ip, "0.0.0.0") == 0)
			sprintf(wan_subnet, "0.0.0.0");
	}

	if (strcmp(lan_subnet, wan_subnet) == 0)
		match[0] = '1';

	sprintf(body, resp, lan_subnet, password, match, lan_mac);
	BuildSendAndCloseSoapResp(req, body);
}

static struct method_struct soapMethod5[] =
{
	{ "SetConfigLANSubnet",  LANConfigSecurity_SetConfigLANSubnet },
	{ "SetConfigPassword", LANConfigSecurity_SetConfigPassword },
	
	{ "GetInfo", LANConfigSecurity_GetInfo },
	
	{ 0, 0 }
};
/**********************************************************/

/****************** WLANConfiguration Methods ******************/
void generate_wepkey(char *phrase, char *len)
{
	char key[4][32] = {"", "", "", ""};
	char *p;
	
	if (strcmp(len, "64") == 0) {
		wep_key_64(phrase, key);

		config_set("key_length", "5");
		config_set("wep_64_key1", key[0]);
		config_set("wep_64_key2",  key[1]);
		config_set("wep_64_key3",  key[2]);
		config_set("wep_64_key4",  key[3]);
		config_set("wl_key1",  key[0]);
		config_set("wl_key2",  key[1]);
		config_set("wl_key3",  key[2]);
		config_set("wl_key4",  key[3]);
	} else {
		p = wep_key_128(phrase);

		config_set("key_length", "13");
		config_set("wep_128_key1", p);
		config_set("wep_128_key2", p);
		config_set("wep_128_key3", p);
		config_set("wep_128_key4", p);
		config_set("wl_key1",  p);
		config_set("wl_key2",  p);
		config_set("wl_key3",  p);
		config_set("wl_key4",  p);
	}
}

void set_region(char *region)
{
	if (strcasecmp(region, "Africa") == 0 || strcasecmp(region, "ZA") == 0)
		config_set("wl_country_code", "1");
	else if (strcasecmp(region, "Asia") == 0  || strcasecmp(region, "SG") == 0)
		config_set("wl_country_code", "2");
	else if (strcasecmp(region, "Australia") == 0  || strcasecmp(region, "AU") == 0)
		config_set("wl_country_code", "3");
	else if (strcasecmp(region, "Canada") == 0  || strcasecmp(region, "CA") == 0)
		config_set("wl_country_code", "5");
	else if (strcasecmp(region, "Europe") == 0  || strcasecmp(region, "GB") == 0)
		config_set("wl_country_code", "7");
	else if (strcasecmp(region, "Israel") == 0  || strcasecmp(region, "IL") == 0)
		config_set("wl_country_code", "9");
	else if (strcasecmp(region, "Japan") == 0  || strcasecmp(region, "JP") == 0)
		config_set("wl_country_code", "10");
	else if (strcasecmp(region, "Korea") == 0  || strcasecmp(region, "KR2") == 0)
		config_set("wl_country_code", "11");
	else if (strcasecmp(region, "Mexico") == 0  || strcasecmp(region, "MX") == 0)
		config_set("wl_country_code", "12");
	else if (strcasecmp(region, "South America") == 0  || strcasecmp(region, "UY") == 0)
		config_set("wl_country_code", "15");
	else
		config_set("wl_country_code", "16");

}

void set_wlmode(char *wl_mode)
{
	if (strcasecmp(wl_mode, "130Mbps") == 0)
		config_set("wl_mode", "2");
	else if (strcasecmp(wl_mode, "270Mbps") == 0)
		config_set("wl_mode", "3");
	else
		config_set("wl_mode", "1");
}

void set_wep_autotype(char *type)
{
	if (strcasecmp(type, "Open") == 0)
		config_set("wl_auth", "2");
	else if (strcasecmp(type, "Shared") == 0)
		config_set("wl_auth", "1");
	else
		config_set("wl_auth", "2");
}

void get_region(char *region)
{
	char buf[8];
	
	sprintf(buf, config_get("wl_country_code"));

	if (strcmp(buf, "0") == 0)
		strcpy(region, "--");
	else if (strcmp(buf, "1") == 0)
		strcpy(region, "Africa");
	else if (strcmp(buf, "2") == 0)
		strcpy(region, "Asia");
	else if (strcmp(buf, "3") == 0)
		strcpy(region, "Australia");
	else if (strcmp(buf, "4") == 0)
		strcpy(region, "Brazil");
	else if (strcmp(buf, "5") == 0)
		strcpy(region, "Canada");
	else if (strcmp(buf, "6") == 0)
		strcpy(region, "China");
	else if (strcmp(buf, "7") == 0)
		strcpy(region, "Europe");
	else if (strcmp(buf, "8") == 0)
		strcpy(region, "India");
	else if (strcmp(buf, "9") == 0)
		strcpy(region, "Israel");
	else if (strcmp(buf, "10") == 0)
		strcpy(region, "Japan");
	else if (strcmp(buf, "11") == 0)
		strcpy(region, "Korea");
	else if (strcmp(buf, "12") == 0)
		strcpy(region, "Mexico");
	else if (strcmp(buf, "13") == 0)
		strcpy(region, "Middle East");
	else if (strcmp(buf, "14") == 0)
		strcpy(region, "Russia");
	else if (strcmp(buf, "15") == 0)
		strcpy(region, "South America");
	else
		strcpy(region, "United States");

}


void get_wlmode(char *wl_mode)
{
	char buf[8];
	
	sprintf(buf, config_get("wl_mode"));

	if (strcmp(buf, "3") == 0)
		strcpy(wl_mode, "Up to 300Mbps");
	else if (strcmp(buf, "2") == 0)
		strcpy(wl_mode, "Up to 145Mbps");
	else
		strcpy(wl_mode, "Up to 54Mbps");
}

static void WLANConfiguration_SetEnable(request * req)
{
	static char resp[] =
		"   <m:SetEnableResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetEnableResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewEnable");

		if (enable && (*enable == '0' || *enable == '1')) {
			if (*enable == '0')
				       config_set("endis_wl_radio","0");
			else
				       config_set("endis_wl_radio","1");

                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");	
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

//Need modify: add config password.
static void WLANConfiguration_SetConfigPassword(request * req)
{
	static char resp[] =
		"   <m:SetConfigPasswordResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetConfigPasswordResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *password;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		password = GetValueFromNameValueList(&data, "NewPassword");

		if (password) {
			
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetChannel(request * req)
{
	static char resp[] =
		"   <m:SetChannelResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetChannelResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *channel;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		channel = GetValueFromNameValueList(&data, "NewChannel");

		if (channel) {
			if (strcasecmp(channel, "AUTO") == 0) {
				config_set("wl_channel", "0");
				config_set("wl_hidden_channel", "0");
			} else {
				config_set("wl_channel", channel);
				config_set("wl_hidden_channel", channel);
			}
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetSSID(request * req)
{
	static char resp[] =
		"   <m:SetSSIDResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetSSIDResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *ssid;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		ssid = GetValueFromNameValueList(&data, "NewSSID");

		if (ssid) {
			config_set("wl_ssid", ssid);
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetSSIDBroadcast(request * req)
{
	static char resp[] =
		"   <m:SetSSIDBroadcastResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetSSIDBroadcastResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewSSIDBroadcast");

		if (enable && (*enable == '0' || *enable == '1')) {
                     if(*enable == '0' )
					 config_set("endis_ssid_broadcast","0");
			else if(*enable == '1')
				        config_set("endis_ssid_broadcast","1");    
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}


static void WLANConfiguration_SetWLANNoSecurity(request * req)
{
	static char resp[] =
		"   <m:SetWLANNoSecurityResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetWLANNoSecurityResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *ssid, *region, *channel, *wl_mode;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		ssid = GetValueFromNameValueList(&data, "NewSSID");
		region = GetValueFromNameValueList(&data, "NewRegion");
		channel = GetValueFromNameValueList(&data, "NewChannel");
		wl_mode = GetValueFromNameValueList(&data, "NewWirelessMode");

		if (ssid && (strcmp(region,"KR2")==0 || strcmp(region,"US")==0 ) && channel && wl_mode) {
			config_set("wl_ssid", ssid);
			config_set("wl_sectype", "1");

			if (strcasecmp(channel, "AUTO") == 0) {
				config_set("wl_channel", "0");
				config_set("wl_hidden_channel", "0");
			} else {
				config_set("wl_channel", channel);
				config_set("wl_hidden_channel", channel);
			}
			
			set_region(region);
			set_wlmode(wl_mode);
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetWLANWEPByKeys(request * req)
{
	static char resp[] =
		"   <m:SetWLANWEPByKeysResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetWLANWEPByKeysResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *ssid, *region, *channel, *wl_mode;
	char *wep_authtype, *weplen, *key[4];

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		ssid = GetValueFromNameValueList(&data, "NewSSID");
		region = GetValueFromNameValueList(&data, "NewRegion");
		channel = GetValueFromNameValueList(&data, "NewChannel");
		wl_mode = GetValueFromNameValueList(&data, "NewWirelessMode");
		wep_authtype = GetValueFromNameValueList(&data, "NewWEPAuthType");
		weplen = GetValueFromNameValueList(&data, "NewWEPLength");
		key[0] = GetValueFromNameValueList(&data, "NewKey1");
		key[1] = GetValueFromNameValueList(&data, "NewKey2");
		key[2] = GetValueFromNameValueList(&data, "NewKey3");
		key[3] = GetValueFromNameValueList(&data, "NewKey4");
		
		if (ssid && (strcmp(region,"KR2")==0 || strcmp(region,"US")==0 ) && channel && wl_mode && wep_authtype && weplen) {
			config_set("wl_ssid", ssid);
			config_set("wl_sectype", "2");
			config_set("wl_key1", key[0]);
			config_set("wl_key2", key[1]);
			config_set("wl_key3", key[2]);
			config_set("wl_key4", key[3]);

			if (strcasecmp(channel, "AUTO") == 0) {
				config_set("wl_channel", "0");
				config_set("wl_hidden_channel", "0");
			} else {
				config_set("wl_channel", channel);
				config_set("wl_hidden_channel", channel);
			}
			
			if (strcmp(weplen, "64") == 0) {
				config_set("key_length", "5");
				config_set("wep_64_key1", key[0]);
				config_set("wep_64_key2", key[1]);
				config_set("wep_64_key3", key[2]);
				config_set("wep_64_key4", key[3]);
			} else {
				config_set("key_length", "13");
				config_set("wep_128_key1", key[0]);
				config_set("wep_128_key2", key[1]);
				config_set("wep_128_key3", key[2]);
				config_set("wep_128_key4", key[3]);
			}
			
			set_region(region);
			set_wlmode(wl_mode);
			set_wep_autotype(wep_authtype);
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetWLANWEPByPassphrase(request * req)
{
	static char resp[] =
		"   <m:SetWLANWEPByPassphraseResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetWLANWEPByPassphraseResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *ssid, *region, *channel, *wl_mode;
	char *wep_authtype, *weplen, *wep_phrase, *wep_key;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		ssid = GetValueFromNameValueList(&data, "NewSSID");
		region = GetValueFromNameValueList(&data, "NewRegion");
		channel = GetValueFromNameValueList(&data, "NewChannel");
		wl_mode = GetValueFromNameValueList(&data, "NewWirelessMode");
		wep_authtype = GetValueFromNameValueList(&data, "NewWEPAuthType");
		weplen = GetValueFromNameValueList(&data, "NewWEPLength");
		wep_phrase = GetValueFromNameValueList(&data, "NewWEPPassphrase");
		
		if (ssid && (strcmp(region,"KR2")==0 || strcmp(region,"US")==0 ) && channel && wl_mode && wep_authtype && weplen && 
		    wep_phrase && strlen(wep_phrase) >= 4 && strlen(wep_phrase) <= 32) {
			config_set("wl_ssid", ssid);
			config_set("wl_sectype", "2");
			config_set("weppassphrase", wep_phrase);

			if (strcasecmp(channel, "AUTO") == 0) {
				config_set("wl_channel", "0");
				config_set("wl_hidden_channel", "0");
			} else {
				config_set("wl_channel", channel);
				config_set("wl_hidden_channel", channel);
			}
			
			set_region(region);
			set_wlmode(wl_mode);
			set_wep_autotype(wep_authtype);
			generate_wepkey(wep_phrase, weplen);
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

static void WLANConfiguration_SetWLANWPAPSKByPassphrase(request * req)
{
	static char resp[] =
		"   <m:SetWLANWPAPSKByPassphraseResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetWLANWPAPSKByPassphraseResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *ssid, *region, *channel, *wl_mode;
	char *wpa_mode, *wpa_phrase;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		ssid = GetValueFromNameValueList(&data, "NewSSID");
		region = GetValueFromNameValueList(&data, "NewRegion");
		channel = GetValueFromNameValueList(&data, "NewChannel");
		wl_mode = GetValueFromNameValueList(&data, "NewWirelessMode");
		wpa_mode = GetValueFromNameValueList(&data, "NewWPAEncryptionModes");
		wpa_phrase = GetValueFromNameValueList(&data, "NewWPAPassphrase");
		
		if (ssid && (strcmp(region,"KR2")==0 || strcmp(region,"US")==0 ) && channel && wl_mode && wpa_mode && wpa_phrase) {
			config_set("wl_ssid", ssid);
			config_set("wl_sectype", "2");

			if (strcasecmp(channel, "AUTO") == 0) {
				config_set("wl_channel", "0");
				config_set("wl_hidden_channel", "0");
			} else {
				config_set("wl_channel", channel);
				config_set("wl_hidden_channel", channel);
			}

			if (strcasecmp(wpa_mode, "WPA-PSK") == 0) {
				config_set("wl_sectype", "3");
				config_set("wl_wpa1_psk", wpa_phrase);
				
			} else if (strcasecmp(wpa_mode, "WPA2-PSK") == 0) {
				config_set("wl_sectype", "4");
				config_set("wl_wpa2_psk", wpa_phrase);
			} else {
				config_set("wl_sectype", "5");
				config_set("wl_wpas_psk", wpa_phrase);
			}
			
			set_region(region);
			set_wlmode(wl_mode);
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

//Need modify: enable & disable WPS.
static void WLANConfiguration_SetWPSMode(request * req)
{
	static char resp[] =
		"   <m:SetWPSModeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"   </m:SetWPSModeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable;

	if (config_state == CONFIG_STARTED) {
		ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
		enable = GetValueFromNameValueList(&data, "NewWPSEnable");

		if (enable && (*enable == '0' || *enable == '1')) {
                     if(*enable == '0' )
					 config_set("wsc_disable","1");
			else if(*enable == '1')
				        config_set("wsc_disable","0");    
			
                     //do_cmd("/usr/sbin/nvram commit");
			do_cmd("/etc/rc.d/wlan.sh restart");
			BuildSendAndCloseSoapResp(req, resp);
		} else
			SoapError(req, "402");

		ClearNameValueList(&data);
	} else
		SoapError(req, "501");
}

//Need modify the value of NewEnable and NewStatus.
static void WLANConfiguration_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewEnable>%s</NewEnable>\r\n"
		"      <NewSSIDBroadcast>%s</NewSSIDBroadcast>\r\n"
		"      <NewStatus>%s</NewStatus>\r\n"
		"      <NewSSID>%s</NewSSID>\r\n"
		"      <NewRegion>%s</NewRegion>\r\n"
		"      <NewChannel>%s</NewChannel>\r\n"
		"      <NewWirelessMode>%s</NewWirelessMode>\r\n"
		"      <NewBasicEncryptionModes>%s</NewBasicEncryptionModes>\r\n"
		"      <NewWEPAuthType>%s</NewWEPAuthType>\r\n"
		"      <NewWPAEncryptionModes>%s</NewWPAEncryptionModes>\r\n"
		"      <NewWLANMACAddress>%s</NewWLANMACAddress>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";
	
	char body[2048], buf[8], buffer[32];
	char enable[8] = "1", status[8] = "Up";
	char ssid_broadcast[8], ssid[32];
	char region[64] = "", channel[8], wl_mode[64] = "", basic_mode[32] = "";
	char wepauth_type[16] = "", wpa_mode[32] = "", wlan_mac[32] = "";

	sprintf(ssid_broadcast, config_get("endis_ssid_broadcast"));
	sprintf(ssid, config_get("wl_ssid"));
	sprintf(channel, config_get("wl_channel"));
	if (strcmp(channel, "0") == 0)
		sprintf(channel, "AUTO");

	get_region(region);
	get_wlmode(wl_mode);
	
	sprintf(buf, config_get("wl_sectype"));
	if (strcmp(buf, "1") == 0) {
		sprintf(basic_mode, "None");
		sprintf(wpa_mode, "None");
	} else if (strcmp(buf, "2") == 0) {
		sprintf(basic_mode, "WEP");
		sprintf(wpa_mode, "None");
	} else if (strcmp(buf, "3") == 0) {
		sprintf(basic_mode, "WPA-PSK");
		sprintf(wpa_mode, "WPA-PSK");
	} else if (strcmp(buf, "4") == 0) {
		sprintf(basic_mode, "WPA2-PSK");
		sprintf(wpa_mode, "WPA2-PSK");
	} else if (strcmp(buf, "5") == 0) {
		sprintf(basic_mode, "WPA-PSK/WPA2-PSK");
		sprintf(wpa_mode, "WPA-PSK/WPA2-PSK");
	}

	sprintf(buf, config_get("wl_auth"));
	if (strcmp(buf, "0") == 0)
		sprintf(wepauth_type, "Open");
	else if (strcmp(buf, "1") == 0)
		sprintf(wepauth_type, "Shared");
	else if (strcmp(buf, "2") == 0)
		sprintf(wepauth_type, "Automatic");

	strcpy(buffer, config_get("lan_hwaddr"));
	change_mac_format(buffer, wlan_mac);

	sprintf(body, resp, enable, ssid_broadcast, status, ssid, region, channel, 
		wl_mode, basic_mode, wepauth_type, wpa_mode, wlan_mac);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetWEPSecurityKeys(request * req)
{
	static char resp[] =
		"   <m:GetWEPSecurityKeysResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewWEPKey>%s</NewWEPKey>\r\n"
		"      <NewWEPPassphrase>%s</NewWEPPassphrase>\r\n"
		"   </m:GetWEPSecurityKeysResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], buf[32];
	char keynum[8], key[32], passphrase[32];

	sprintf(keynum, config_get("wl_key"));
	sprintf(buf, "wl_key%s", keynum);
	if(config_get(buf))
		strcpy(key, config_get(buf));
	else
		strcpy(key, "");
		
	if(config_get("weppassphrase"))	
		sprintf(passphrase, config_get("weppassphrase"));
	else
		strcpy(passphrase, "");

	sprintf(body, resp, key, passphrase);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetWPASecurityKeys(request * req)
{
	static char resp[] =
		"   <m:GetWPASecurityKeysResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewWPAPassphrase>%s</NewWPAPassphrase>\r\n"
		"   </m:GetWPASecurityKeysResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], type[8], passphrase[64];

	sprintf(type, config_get("wl_sectype"));
	if (strcmp(type, "3") == 0)
		sprintf(passphrase, config_get("wl_wpa1_psk"));
	else if (strcmp(type, "4") == 0)
		sprintf(passphrase, config_get("wl_wpa2_psk"));
	else
		sprintf(passphrase, config_get("wl_wpas_psk"));

	sprintf(body, resp, passphrase);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetSSID(request * req)
{
	static char resp[] =
		"   <m:GetSSIDResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewSSID>%s</NewSSID>\r\n"
		"   </m:GetSSIDResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], ssid[128];

	sprintf(ssid, config_get("wl_ssid"));

	sprintf(body, resp, ssid);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetChannelInfo(request * req)
{
	static char resp[] =
		"   <m:GetChannelInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewChannel>%s</NewChannel>\r\n"
		"   </m:GetChannelInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], channel[8];

	sprintf(channel, config_get("wl_channel"));
	if (strcmp(channel, "0") == 0)
		sprintf(channel, "AUTO");

	sprintf(body, resp, channel);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetRegion(request * req)
{
	static char resp[] =
		"   <m:GetRegionResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewRegion>%s</NewRegion>\r\n"
		"   </m:GetRegionResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], region[64] = "";

	get_region(region);

	sprintf(body, resp, region);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetWirelessMode(request * req)
{
	static char resp[] =
		"   <m:GetWirelessModeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewWirelessMode>%s</NewWirelessMode>\r\n"
		"   </m:GetWirelessModeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], wl_mode[64];

	get_wlmode(wl_mode);

	sprintf(body, resp, wl_mode);
	BuildSendAndCloseSoapResp(req, body);
}

static void WLANConfiguration_GetSSIDBroadcast(request * req)
{
	static char resp[] =
		"   <m:GetSSIDBroadcastResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewSSIDBroadcast>%s</NewSSIDBroadcast>\r\n"
		"   </m:GetSSIDBroadcastResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], enable[8];

	sprintf(enable, config_get("endis_ssid_broadcast"));

	sprintf(body, resp, enable);
	BuildSendAndCloseSoapResp(req, body);
}

//Need modify the value of NewWPSEnable.
static void WLANConfiguration_GetWPSMode(request * req)
{
	static char resp[] =
		"   <m:GetWPSModeResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewWPSEnable>%s</NewWPSEnable>\r\n"
		"   </m:GetWPSModeResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024];
	char enable[8] ;
	sprintf(enable,config_get("wsc_disable"));
	if(*enable == '0')
		*enable='1';
	else if (*enable == '1')
		*enable='0';

	sprintf(body, resp, enable);
	BuildSendAndCloseSoapResp(req, body);
}

//Need modify the value of NewWPSPINEnable and NewRouterWPSPIN.
static void WLANConfiguration_GetWPSPINInfo(request * req)
{
	static char resp[] =
		"   <m:GetWPSPINInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:WLANConfiguration:1\">\r\n"
		"      <NewWPSPINEnable>%s</NewWPSPINEnable>\r\n"
		"      <NewRouterWPSPIN>%s</NewRouterWPSPIN>\r\n"
		"   </m:GetWPSPINInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024];
	char enable[8], pin[32] ;
	sprintf(enable,config_get("endis_pin"));
	sprintf(pin,config_get("wps_pin"));

       if(*enable == '0')
		*enable='1';
	else if (*enable == '1')
		*enable='0';

	  
	sprintf(body, resp, enable, pin);
	BuildSendAndCloseSoapResp(req, body);
}

static struct method_struct soapMethod6[] =
{
	{ "SetEnable",  WLANConfiguration_SetEnable },
	{ "SetConfigPassword", WLANConfiguration_SetConfigPassword },
	{ "SetChannel", WLANConfiguration_SetChannel },
	{ "SetSSIDBroadcast", WLANConfiguration_SetSSIDBroadcast },
	{ "SetSSID",  WLANConfiguration_SetSSID },
	{ "SetWLANNoSecurity", WLANConfiguration_SetWLANNoSecurity },
	{ "SetWLANWEPByKeys", WLANConfiguration_SetWLANWEPByKeys },
	{ "SetWLANWEPByPassphrase",  WLANConfiguration_SetWLANWEPByPassphrase },
	{ "SetWLANWPAPSKByPassphrase", WLANConfiguration_SetWLANWPAPSKByPassphrase },
	{ "SetWPSMode", WLANConfiguration_SetWPSMode },
	
	{ "GetInfo", WLANConfiguration_GetInfo },
	{ "GetSSIDBroadcast",  WLANConfiguration_GetSSIDBroadcast },
	{ "GetWEPSecurityKeys",  WLANConfiguration_GetWEPSecurityKeys },
	{ "GetWPASecurityKeys", WLANConfiguration_GetWPASecurityKeys },
	{ "GetSSID", WLANConfiguration_GetSSID },
	{ "GetChannelInfo",  WLANConfiguration_GetChannelInfo },
	{ "GetRegion", WLANConfiguration_GetRegion },
	{ "GetWirelessMode", WLANConfiguration_GetWirelessMode },
	{ "GetWPSMode", WLANConfiguration_GetWPSMode },
	{ "GetWPSPINInfo", WLANConfiguration_GetWPSPINInfo },
	
	{ 0, 0 }
};
/**********************************************************/

/********************** Time Methods ************************/
static void Time_GetInfo(request * req)
{
	static char resp[] =
		"   <m:GetInfoResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service: Time:1\">\r\n"
		"      <NewNTPServer1>%s</NewNTPServer1>\r\n"
		"      <NewNTPServer2>%s</NewNTPServer2>\r\n"
		"      <NewNTPServer3>%s</NewNTPServer3>\r\n"
		"      <NewNTPServer4>%s</NewNTPServer4>\r\n"
		"   </m:GetInfoResponse>\r\n"
		"   <ResponseCode>000</ResponseCode>\r\n";

	char body[1024], ntpserver1[64] = "", ntpserver2[64] = "";

	strcpy(ntpserver1, config_get("ntpserver1"));
	strcpy(ntpserver2, config_get("ntpserver2"));

	sprintf(body, resp, ntpserver1, ntpserver2, ntpserver1, ntpserver2);
	BuildSendAndCloseSoapResp(req, body);
}

static struct method_struct soapMethod7[] =
{
	{ "GetInfo", Time_GetInfo },
		
	{ 0, 0 }
};
/**********************************************************/


/*************** ParentalControl Methods ******************/
#define LEN_MACADDR 12
#define LEN_DEVICEID 32
#define MAX_BUF 128
#define PC_TABLE "/tmp/ParentalControl/ParentalControl_table"
#define HANDLE_PC_TABLE_PUT  "/sbin/handle_pc_table put"
#define HANDLE_PC_TABLE_GET  "/sbin/handle_pc_table get"

typedef struct ParentalControl_map
{
	char macaddr[LEN_MACADDR+1];
	char deviceid[LEN_DEVICEID+1];
	struct ParentalControl_map * next;
}PC_map, *PC_map_p;

/******************************************
 * * Read all mac addresses and deviceID records from ParentalControl_table,
 * * and save these as a linklist...
 * * Parameters:	first -- an out parameter, save the first link pointer of the linklist
 * *		last -- an out parameter, save the last link pointer of the linklist
 * * Return:	-1 -- error
 * * 		0 -- there is no any records in the file ParentalControl_table.
 * * 		[1,32] -- the count number of the records have read. 
 ******************************************/
static int get_all_macaddr(PC_map_p * first, PC_map_p * last)
{
	FILE *fp=NULL;
	char buf[MAX_BUF]={0};
	PC_map_p tmp={0};
	int record_num=0;	

	if(!(fp=fopen(PC_TABLE,"r")))
	{		
		return (0);
	}
	while(fgets(buf, MAX_BUF, fp))
	{
		if(!(tmp=(PC_map_p)malloc(sizeof(PC_map))))
		{			
			return (-1);
		}
		sscanf(buf, "%s %s", tmp->macaddr, tmp->deviceid);
		
		if(!(*first))
		{
			tmp->next=NULL;
			*first=tmp;
		}
		else
		{
			tmp->next=NULL;
			(*last)->next=tmp;
		}

		*last=tmp;
		record_num++;	
	}
	
	if(fp)
		fclose(fp);
	return record_num; 
}

/*****************************************
 * * free ram space of linklist
 ****************************************/
static void free_linklist(PC_map_p * first)
{
	PC_map_p tmp={0};
	while(*first)
	{
		tmp=*first;
		*first=(*first)->next;
		free(tmp);
	}
}

/*****************************************
 * *write all mac address and deviceID records to the file ParentalControl_table
 * * parameters:	first -- the pointer of linklist that track records of macaddr 
 * * 			deviceid informations.
 * * return:	0 -- write successful.
 * * 		other -- write failed.
 *****************************************/
static int write_all_macaddr(PC_map_p * first)
{
	FILE * fp=NULL;
	char buf[MAX_BUF]={0};
	PC_map_p tmp={0};
	
	if(!(fp=fopen(PC_TABLE, "w")))
	{		
		return (-1);
	}	

	tmp=*first;
	while(tmp)
	{
		memset(buf, 0, MAX_BUF);
		sprintf(buf, "%s %s", tmp->macaddr, tmp->deviceid);
		strcat(buf, "\r\n");
                fputs(buf, fp);
                tmp=tmp->next;
	}

	if(fp)
		fclose(fp);
	system(HANDLE_PC_TABLE_PUT);
	system("/usr/sbin/nvram commit");
	return 0;	
}

static void PC_GetDNSMasqDeviceID(request * req)
{
	static char resp[] =
		"   <m:GetDNSMasqDeviceIDResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"      <NewDeviceID>%s</NewDeviceID>\r\n"
		"   </m:GetDNSMasqDeviceIDResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	char body[1024]={0}, deviceid[LEN_DEVICEID+1]={0};
	struct NameValueParserData data;
	char *macaddr;
	PC_map_p first_record={0}, last_record={0}, tmp_record={0};

	ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
	macaddr = GetValueFromNameValueList(&data, "NewMACAddress");
	if((12 != strlen(macaddr)) && strcasecmp("default", macaddr))
	{
	        SoapError(req, "1");
                ClearNameValueList(&data);
                return ;
	}

	get_all_macaddr(&first_record, &last_record);
	tmp_record=first_record;
	while(tmp_record)
	{
		if(!strcasecmp(macaddr, tmp_record->macaddr))
		{
			strcpy(deviceid, tmp_record->deviceid);
			break;
		}
		tmp_record=tmp_record->next;
	}
	
	if(!strcmp(deviceid, ""))
	{
			SoapError(req, "1");
			ClearNameValueList(&data);
			free_linklist(&first_record);
			return ;
	}

	sprintf(body, resp, deviceid);
	BuildSendAndCloseSoapResp(req, body);
	ClearNameValueList(&data);
	free_linklist(&first_record);
}

static void PC_SetDNSMasqDeviceID(request * req)
{
	static char resp[] =
		"   <m:SetDNSMasqDeviceIDResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"   </m:SetDNSMasqDeviceIDResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *macaddr, *deviceid;
	char buf[MAX_BUF]={0};
	PC_map_p first_record={0}, last_record={0}, tmp_record={0};
	int record_num=0;

	ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
	macaddr = GetValueFromNameValueList(&data, "NewMACAddress");
	deviceid = GetValueFromNameValueList(&data, "NewDeviceID");
	if(((12 != strlen(macaddr)) && strcasecmp("default", macaddr)) || (16 != strlen(deviceid)))
	{
                SoapError(req, "2");
                ClearNameValueList(&data);
                free_linklist(&first_record);
                return;
	}

	record_num = get_all_macaddr(&first_record, &last_record);
	if(0 > record_num)
	{
	        SoapError(req, "2");
                ClearNameValueList(&data);
                free_linklist(&first_record);
                return;	
	}

	tmp_record=first_record;
	while(tmp_record)
	{
		if(!strcasecmp(tmp_record->macaddr, macaddr))
		{
			strcpy(tmp_record->deviceid, deviceid);
			break;
		}
		tmp_record=tmp_record->next;
	}
	if(!tmp_record)
	{
		if(32 <= record_num)
		{
			SoapError(req, "1");
			ClearNameValueList(&data);
			free_linklist(&first_record);
			return;
		}

	
		if(!(tmp_record=(PC_map_p)malloc(sizeof(PC_map))))
		{
			SoapError(req, "2");
			ClearNameValueList(&data);
			free_linklist(&first_record);
			return;
		}
		else
		{
			strcpy(tmp_record->macaddr, macaddr);
			strcpy(tmp_record->deviceid, deviceid);
			tmp_record->next=NULL;
			if(last_record)
			{
				last_record->next=tmp_record;
				last_record=last_record->next;
			}
			else
			{
				first_record=last_record=tmp_record;	
			}
		}
	}	
	if(write_all_macaddr(&first_record))
                SoapError(req, "2");
	else	
	{
		BuildSendAndCloseSoapResp(req, resp);		
	}
	ClearNameValueList(&data);
	free_linklist(&first_record);
}

static void PC_EnableParentalControl(request * req)
{
	static char resp[] =
		"   <m:EnableParentalControlResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"   </m:EnableParentalControlResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *enable;

	ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
	enable = GetValueFromNameValueList(&data, "NewEnable");

	if (enable && (*enable == '0' || *enable == '1')) 
	{
		if(!config_set("ParentalControl", enable))
		{
			if(*enable == '1')
			{
				system("echo \"nameserver 208.67.222.222\" > /tmp/resolv.conf");
				system("echo \"nameserver 208.67.220.220\" >> /tmp/resolv.conf");
				config_set("wan_ether_dns_assign", "1");
				config_set("wan_ether_dns1", "208.67.222.222");
				config_set("wan_ether_dns2", "208.67.220.220");
			}
			else
			{
				config_set("wan_ether_dns_assign", "0");
				system("/sbin/cmdwan restart");
			}
			system("/usr/sbin/nvram commit");
			BuildSendAndCloseSoapResp(req, resp);
		}
		else
		{
			SoapError(req, "1");
		}
	} 
	else
	{
		SoapError(req, "1");
	}
	ClearNameValueList(&data);
}

static void PC_GetAllMACAddresses(request * req)
{
	static char resp[] =
		"   <m:GetAllMACAddressesResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"      <AllMACAddresses>%s</AllMACAddresses>\r\n"
		"   </m:GetAllMACAddressesResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	char body[1024]={0};
	char buf[600]={0};
	PC_map_p first_record={0}, last_record={0}, tmp_record={0};

	if(0 > get_all_macaddr(&first_record, &last_record))
	{
		SoapError(req, "1");
		free_linklist(&first_record);
		return ;
	}
	tmp_record=first_record;

	while(tmp_record)
	{
		strcat(buf, tmp_record->macaddr);
		strcat(buf, "|");
		tmp_record=tmp_record->next;
	}
	if('|' == buf[strlen(buf)-1])
		buf[strlen(buf)-1]='\0';

	sprintf(body, resp, buf);
	BuildSendAndCloseSoapResp(req, body);
	free_linklist(&first_record);
}

static void PC_DeleteMACAddress(request * req)
{
	static char resp[] =
		"   <m:DeleteMACAddressResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"   </m:DeleteMACAddressResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	struct NameValueParserData data;
	char *macaddr;
	PC_map_p first_record={0}, last_record={0}, tmp_record={0}, now_record={0};
	int  success=0;

	ParseNameValue(req->header_line, atoi(req->content_length) + 2, &data);
	macaddr = GetValueFromNameValueList(&data, "NewMACAddress");
	if((12 != strlen(macaddr)) && strcasecmp("default", macaddr))
	{
                SoapError(req, "1");
	        ClearNameValueList(&data);	
		return;
	}

	if(0 > get_all_macaddr(&first_record, &last_record))
	{
		SoapError(req, "1");
		free_linklist(&first_record);
		return ;
	}
	tmp_record=now_record=first_record;
	while(tmp_record)
	{
		if(!strcasecmp(tmp_record->macaddr, macaddr))
		{
			if(tmp_record == first_record)
			{
				first_record=tmp_record->next;
				free(tmp_record);
			}
			else
			{
				now_record->next=tmp_record->next;
				free(tmp_record);	
			}
			success=1;
			break;
		}
		now_record=tmp_record;
		tmp_record=tmp_record->next;
	}
	
	if(success && !(write_all_macaddr(&first_record)))
	{
		BuildSendAndCloseSoapResp(req, resp);		
	}
	else
		SoapError(req, "1");
	ClearNameValueList(&data);
	free_linklist(&first_record);
}

static void PC_GetEnableStatus(request * req)
{
	static char resp[] =
		"   <m:GetEnableStatusResponse\r\n"
		"      xmlns:m=\"urn:NETGEAR-ROUTER:service:ParentalControl:1\">\r\n"
		"      <ParentalControl>%s</ParentalControl>\r\n"
		"   </m:GetEnableStatusResponse>\r\n"
		"   <ResponseCode>0</ResponseCode>\r\n";

	char *enable;
	char body[1024]={0};

	enable=config_get("ParentalControl");
	if(strcmp(enable, "0") && strcmp(enable, "1"))
	{
		SoapError(req, "1");
		return;
	}

	sprintf(body, resp, enable);
	BuildSendAndCloseSoapResp(req, body);
}


static struct method_struct soapMethod8[] =
{
	{ "GetDNSMasqDeviceID", PC_GetDNSMasqDeviceID },
	{ "SetDNSMasqDeviceID", PC_SetDNSMasqDeviceID },
	{ "EnableParentalControl", PC_EnableParentalControl },
	{ "GetAllMACAddresses", PC_GetAllMACAddresses },
	{ "DeleteMACAddress", PC_DeleteMACAddress },
	{ "GetEnableStatus", PC_GetEnableStatus },	
	{ 0, 0 }
};
/**********************************************************/


int ExecuteSoapAction(request *req, char *action, int n)
{
	int len;
	char *p, *q;
	struct method_struct *soap;

	/* Try to read more info */
	int bytes, bytes_left;

	/* reset to block socket */
	struct timeval timo;
	fcntl(req->fd, F_SETFL, 0);
	timo.tv_sec = 0;
	timo.tv_usec = 500000; // 500 ms
	setsockopt(req->fd, SOL_SOCKET, SO_SNDTIMEO, &timo, sizeof(timo));
	setsockopt(req->fd, SOL_SOCKET, SO_RCVTIMEO, &timo, sizeof(timo));

	bytes_left = CLIENT_STREAM_SIZE - req->client_stream_pos;
	bytes = read(req->fd, req->client_stream + req->client_stream_pos, bytes_left);
	while (bytes > 0) {
		req->client_stream_pos += bytes;
		bytes_left = CLIENT_STREAM_SIZE - req->client_stream_pos;
		bytes = read(req->fd, req->client_stream + req->client_stream_pos, bytes_left);
	}

	p = strstr(action, ":service:");

	if (p == NULL)	goto err;
	
	p = p + 9;
	if (p == NULL)	goto err;

	q = p;
	len = 0;
	while (1) {
		if (*q == ':' || *q == '\0')	break;
		
		q++; len++;
	}

	if (*q != ':')	goto err;

	if (strncmp(p, "DeviceConfig", len) == 0)
		soap = &soapMethod1[0];
	else if (strncmp(p, "DeviceInfo", len) == 0)
		soap = &soapMethod2[0];
	else if (strncmp(p, "WANIPConnection", len) == 0)
		soap = &soapMethod3[0];
	else if (strncmp(p, "WANEthernetLinkConfig", len) == 0)
		soap = &soapMethod4[0];
	else if (strncmp(p, "LANConfigSecurity", len) == 0)
		soap = &soapMethod5[0];
	else if (strncmp(p, "WLANConfiguration", len) == 0)
		soap = &soapMethod6[0];
	else if (strncmp(p, "Time", len) == 0)
		soap = &soapMethod7[0];
	else if (strncmp(p, "ParentalControl", len) == 0)
		soap = &soapMethod8[0];
	else
		goto err;

	p = strchr(q, '#');

	if (p) {
		p++;
		while(soap->methodName) {
			len = strlen(soap->methodName);
			if (strncmp(p, soap->methodName, len) == 0) {
				soap->methodImpl(req);
				return 1;
			}
			soap++;
		}
	}

err:
	SoapError(req, "401");
	return 0;
}

void SoapError(request *req, char *err)
{
	char resp[1024];

	sprintf(resp, "   <ResponseCode>%s</ResponseCode>\r\n", err);
	
	BuildSendAndCloseSoapResp(req, resp);
}
